#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
## batteries
import os
import sys
import unittest
## 3rd party
import pandas as pd
## package
from leylab_pipelines import Dilute
from leylab_pipelines import Utils


# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
class Test_Dilute_import(unittest.TestCase):

    def setUp(self):
        concfile = os.path.join(data_dir, 'conc_file1.txt')
        args = Dilute.parse_args([concfile])
        Dilute.check_args(args)
        self.df_conc = Dilute.conc2df(args.concfile)

    def tearDown(self):
        self.df_conc = None
        pass

    # import
    def test_load_conc_txt(self):
        self.assertTrue(isinstance(self.df_conc, pd.DataFrame))


class Test_Dilute_rows(unittest.TestCase):

    def setUp(self):
        concfile = os.path.join(data_dir, 'conc_file1.txt')
        self.args = Dilute.parse_args(['--rows', '1,2,4-5', concfile])
        Dilute.check_args(self.args)
        self.df_conc = Dilute.conc2df(self.args.concfile,
                                     row_select=self.args.rows)

    def tearDown(self):
        self.df_conc = None
        pass

    # import
    def test_load_conc_txt(self):
        self.assertTrue(isinstance(self.df_conc, pd.DataFrame))
        self.assertEqual(self.df_conc.shape[0], 4)
        self.assertListEqual(list(self.df_conc.index), [0,1,3,4])

class Test_Dilute_check_df_conc(unittest.TestCase):

    def setUp(self):
        concfile = os.path.join(data_dir, 'conc_file1.txt')
        self.args = Dilute.parse_args([concfile])
        Dilute.check_args(self.args)
        self.df_conc = Dilute.conc2df(self.args.concfile)

    def tearDown(self):
        self.df_conc = None
        pass

    # import
    def test_load_conc_txt(self):
        self.assertTrue(isinstance(self.df_conc, pd.DataFrame))
        Dilute.check_df_conc(self.df_conc, self.args)

class Test_Dilute_dil_vols(unittest.TestCase):

    def setUp(self):
        concfile = os.path.join(data_dir, 'conc_file1.txt')
        self.args = Dilute.parse_args([concfile])
        Dilute.check_args(self.args)
        self.df_conc = Dilute.conc2df(self.args.concfile)
        Dilute.check_df_conc(self.df_conc, self.args)
        self.df_conc = Dilute.dilution_volumes(self.df_conc,
                                               dilute_conc=self.args.dilution,
                                               min_vol=self.args.minvolume,
                                               max_vol=self.args.maxvolume,
                                               min_total=self.args.mintotal,
                                               dest_type=self.args.desttype)

    def tearDown(self):
        self.df_conc = None
        pass

    # import
    def test_load_conc_txt(self):
        self.assertTrue(isinstance(self.df_conc, pd.DataFrame))


class Test_Dilute_addDest(unittest.TestCase):

    def setUp(self):
        concfile = os.path.join(data_dir, 'conc_file1.txt')
        self.args = Dilute.parse_args([concfile])
        Dilute.check_args(self.args)
        self.df_conc = Dilute.conc2df(self.args.concfile)

    def tearDown(self):
        self.df_conc = None
        self.args = None

    # adding destination
    def test_load_conc_txt(self):
        self.df_conc = Dilute.add_dest(self.df_conc, self.args.dest, self.args.deststart)
        self.assertTrue(isinstance(self.df_conc, pd.DataFrame))

    # destination start offset
    def test_load_conc_deststart(self):
        self.df_conc = Dilute.add_dest(self.df_conc,
                                       self.args.dest,
                                       dest_start=49)

        self.assertTrue(isinstance(self.df_conc, pd.DataFrame))
        loc_start = self.df_conc.loc[0,'dest_location']
        self.assertEqual(loc_start, 49)


class Test_Dilute_main1(unittest.TestCase):

    def setUp(self):
        concfile = os.path.join(data_dir, 'conc_file1.txt')
        self.args = Dilute.parse_args(['--prefix', '/tmp/DIL1', concfile])
        self.files = Dilute.main(self.args)

    def tearDown(self):
        pass

    def test_main_gwl(self):
        ret = Utils.check_gwl(self.files[0])
        self.assertIsNone(ret)


class Test_Dilute_main2(unittest.TestCase):

    def setUp(self):
        concfile = os.path.join(data_dir, 'conc_file2.txt')
        self.args = Dilute.parse_args(['--prefix', '/tmp/DIL2', concfile])
        self.files = Dilute.main(self.args)

    def tearDown(self):
        pass

    def test_main_gwl(self):
        ret = Utils.check_gwl(self.files[0])
        self.assertIsNone(ret)



# -*- coding: utf-8 -*-

# import
from __future__ import print_function
import os
import sys
import numpy as np


def xstr(x):
    if x is None:
        return ''
    else:
        return x

def _psbl_liq_cls():
    """Returns a set of possible liquid classes available for Fluent
    """
    x = ('Water Free Multi', 'Water Free Single',
         'MasterMix Free Multi', 'MasterMix Free Single', 
         'Ethanol Free Multi', 'Ethanol Free Single', 
         'DMSO Free Multi', 'DMSO Free Single', 
         'Serum Free Multi', 'Serum Free Single',
         'Water Contact Wet Multi', 'Water Contact Wet Single',
         'Water Mix')
    return x

class asp_disp():
    """Commands for aliquoting mastermix
    *Parameters*
    RackLabel
    RackID
    RackType
    Position
    TubeID
    Volume
    LiquidClass 
    TipType
    TipMask
    ForceRack
    MinDetected
    """

    def __init__(self):
        self._ID = ''
        # aspirate parameters
        self.RackLabel = None
        self.RackID = None
        self.RackType = None
        self._Position = 1
        self.TubeID = None
        self.Volume = None
        self._LiquidClass = 'Water Free Single'
        self.TipType = None
        self.TipMask = None
        self.key_order = ['_ID',
                          'RackLabel', 'RackID', 'RackType',
                          'Position', 'TubeID', 'Volume',
                          'LiquidClass', 'TipType', 'TipMask']
        self.psbl_liq_cls = _psbl_liq_cls()


    def cmd(self):
        # list of values in correct order
        vals = [getattr(self, x) for x in self.key_order]
        # None to blank string
        vals = [xstr(x) for x in vals]
        # convert all to strings
        vals = [str(x) for x in vals]
        # return
        return ';'.join(vals)

    def liquid_classes(self):
        x = '\n,'.join(list(self.psbl_liq_cls))
        print(x)        

    @property
    def Position(self):
        return self._Position

    @Position.setter
    def Position(self, value):
        self._Position = int(value)

    @property
    def LiquidClass(self):
        return self._LiquidClass

    @LiquidClass.setter
    def LiquidClass(self, value):
        if value not in self.psbl_liq_cls:
            msg = 'Liquid class "{}" not allowed'
            raise TypeError(msg.format(value))
        self._LiquidClass = value



class aspirate(asp_disp):
    def __init__(self):
        asp_disp.__init__(self)
        self._ID = 'A'


class dispense(asp_disp):
    def __init__(self):
        asp_disp.__init__(self)
        self._ID = 'D'


class multi_disp():
    """Commands for aliquoting mastermix
    *AspirateParameters*
    SrcRackLabel
    SrcRackID
    SrcRackType
    SrcPosition
    *DispenseParameters*
    DestRackLabel
    DestRackID
    DestRackType
    DestPosStart
    DestPosEnd
    *Samples*
    SampleCount
    *Other*
    Volume = How much volume per dispense?
    LiquidClass = Which liquid class to use? Default: 'Water Free Multi'
    NoOfMultiDisp = How many multi-dispenses?
    
    Returns
    * string of commands
    """

    def __init__(self):
        self._ID = 'R'
        # aspirate parameters
        self.SrcRackLabel = None
        self.SrcRackID = None
        self.SrcRackType = None
        self.SrcPosition = 1
        # dispense parameters
        self.DestRackLabel = []
        self.DestRackID = []
        self.DestRackType = []
        self._DestPositions = [1]
        # other
        self.Volume = 1
        self._LiquidClass = 'Water Free Multi'
        self.NoOfMultiDisp = 2
        self.psbl_liq_cls = _psbl_liq_cls()

    def xstr(self, x):
        if x is None:
            return ''
        else:
            return x

    def cmd(self):
        # volume as interable
        if hasattr(self.Volume, '__iter__'):
            self.Volumes = self.Volume
        else:
            self.Volumes = [self.Volume] * len(self.DestPositions)
        # each multi-disp
        steps = []
        for i in range(0, self.SampleCount, self.NoOfMultiDisp):
            # number of dispenses
            if self.SampleCount - i < self.NoOfMultiDisp:
                n_disp = self.SampleCount - i
            else:
                n_disp = self.NoOfMultiDisp
            # amount to aspirate
            asp_vol = round(self.Volume * n_disp, 2)
            # single-asp
            asp = aspirate()
            asp.RackLabel = self.SrcRackLabel
            asp.Position = self.SrcPosition
            asp.Volume = sum(self.Volumes[i:(i+n_disp)])   #asp_vol
            asp.Volume = round(asp.Volume, 2)
            asp.LiquidClass = self.LiquidClass
            steps.append(asp.cmd())
            # multi-disp
            for ii in range(n_disp):
                disp = dispense()
                disp.RackLabel = self.DestRackLabel[i+ii]
                disp.Position = self.DestPositions[i+ii]
                disp.Volume = round(self.Volumes[i+ii], 2)
                disp.LiquidClass = self.LiquidClass
                steps.append(disp.cmd())
            steps.append('W;')
                
        return '\n'.join(steps)

    @property
    def LiquidClass(self):
        return self._LiquidClass

    @LiquidClass.setter
    def LiquidClass(self, value):
        if value not in self.psbl_liq_cls:
            msg = 'Liquid class "{}" not allowed'
            raise TypeError(msg.format(value))
        self._LiquidClass = value

    @property
    def DestPositions(self):
        return self._DestPositions

    @DestPositions.setter
    def DestPositions(self, values):
        try:
            values = values.tolist()
        except AttributeError:
            pass
        values = [int(x) for x in values]
        assert min(values) > 0, 'Min position is 1'
        self._DestPositions = values 

    @property
    def SampleCount(self):
        return len(self._DestPositions)



class reagent_distribution():
    """Commands for aliquoting mastermix
    *AspirateParameters*
    SrcRackLabel
    SrcRackID
    SrcRackType
    SrcPosStart
    SrcPosEnd
    *DispenseParameters*
    DestRackLabel
    DestRackID
    DestRackType
    DestPosStart
    DestPosEnd
    *Other*
    Volume = How much volume to asp/disp?
    LiquidClass = Which liquid class to use? Default: 'Water Free Multi'
    NoOfDiTiReuses = How many times to reuse tips?
    NoOfMultiDisp = How many multi-dispenses?
    Direction = Which way to pipette? Default:0
    ExcludedDestWell = Semi-colon separated string of locations to exclude

    *WashParameters*
    None?

    # Example: R;100ml_2;;Trough 100ml;1;1;96 Well Skirted PCR[003];;96 Well Skirted PCR;1;96;20;Water Free Multi;1;5;0
    """

    def __init__(self, ):
        self._ID = 'R'
        # aspirate parameters
        self.SrcRackLabel = None
        self.SrcRackID = None
        self.SrcRackType = None
        self.SrcPosStart = 1
        self.SrcPosEnd = 1
        # dispense parameters
        self.DestRackLabel = None
        self.DestRackID = None
        self.DestRackType = None
        self.DestPositions = 1
        self.DestPosEnd = 1
        # other
        self.Volume = 1
        self.LiquidClass = 'Water Free Multi'
        self.NoOfDiTiReuses = 1
        self.NoOfMultiDisp = 5
        self.Direction = 0
        self.ExcludedDestWell = None
        self.key_order = ['_ID',
                          'SrcRackLabel', 'SrcRackID', 'SrcRackType',
                          'SrcPosStart', 'SrcPosEnd',
                          'DestRackLabel', 'DestRackID', 'DestRackType',
                          'DestPosStart', 'DestPosEnd',
                          'Volume', 'LiquidClass', 'NoOfDiTiReuses',
                          'NoOfMultiDisp', 'Direction', 'ExcludedDestWell']

    def xstr(self, x):
        if x is None:
            return ''
        else:
            return x

    def cmd(self):
        # list of values in correct order
        vals = [getattr(self, x) for x in self.key_order]
        # None to blank string
        vals = [xstr(x) for x in vals]
        # convert all to strings
        vals = [str(x) for x in vals]
        # return
        return ';'.join(vals)



# main
if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
## batteries
import os
import sys
import unittest
## 3rd party
import pandas as pd
## package
from leylab_pipelines import Utils


# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
class Test_Utils(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_make_range(self):
        x = Utils.make_range('all')
        self.assertIsNone(x)

        x = Utils.make_range('0')
        self.assertListEqual(x, [0])

        x = Utils.make_range('1,2,5')
        self.assertListEqual(x, [1,2,5])

        x = Utils.make_range('1,2,5-6')
        self.assertListEqual(x, [1,2,5,6])

        x = Utils.make_range('1-3,5-6')
        self.assertListEqual(x, [1,2,3,5,6])


    def test_make_range_zeroindex(self):
        x = Utils.make_range('all', set_zero_index=True)
        self.assertIsNone(x)

        with self.assertRaises(ValueError):
            Utils.make_range('0', set_zero_index=True)

        x = Utils.make_range('1,2,5', set_zero_index=True)
        self.assertListEqual(x, [0,1,4])

        x = Utils.make_range('1,2,5-6', set_zero_index=True)
        self.assertListEqual(x, [0,1,4,5])

    def test_check_gwl(self):
        gwl_file = os.path.join(data_dir, 'multi_dispense.gwl')
        ret = Utils.check_gwl(gwl_file)
        self.assertIsNone(ret)



# -*- coding: utf-8 -*-
# import
## batteries
from __future__ import print_function
import os
import sys
import argparse
from itertools import product
## 3rd party
import numpy as np
import pandas as pd
## package
from leylab_pipelines import Utils
from leylab_pipelines import Fluent


# functions
def get_desc():
    desc = 'Convert a mapping file to a NGS amplicon worklist file for the TECAN robot'
    return desc

def parse_args(test_args=None, subparsers=None):
    # desc
    #desc = 'Convert a mapping file to a NGS amplicon worklist file for the TECAN robot'
    desc = get_desc
    epi = """DESCRIPTION:
    Convert a QIIME-formatted mapping file to a GWL file, which is used by the TECAN
    robot to conduct the NGS amplicon PCR prep (ie., combining MasterMix, primers, samples, etc).
    The extra columns in the mapping file designate the SOURCE of samples and primers;
    the DESTINATION (plate & well) is set by this script. 

    EXTRA COLUMNS in MAPPING FILE:
    * "TECAN_sample_labware" = The sample labware name on the robot worktable
    * "TECAN_sample_location" = The well or tube location (a number)
    * "TECAN_primer_labware" = The primer plate labware name on the robot worktable
    * "TECAN_primer_location" = The well location (1-96 or 1-384)
    * "TECAN_sample_rxn_volume" = The volume of sample to use per PCR (ul)

    CONTROLS:
    * For the positive & negative controls, include them in the mapping file.
    * If the controls (or samples) are provided in a tube, use "micro15[XXX]" 
      for the "TECAN_sample_labware" column, but change "XXX" to the tube number
      that you want to use (eg., micro15[003] for tube position 3)

    OUTPUT FILES:
    * The output files ending in "_win" have Windows line breads (needed for the robot)

    MISC NOTES:
    * All volumes are in ul
    * Plate well locations are 1 to n-wells; numbering by column
    * PicoGreen should be added to the MasterMix *prior* to loading on robot
    """
    if subparsers:
        parser = subparsers.add_parser('map2robot', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # args
    ## I/O
    groupIO = parser.add_argument_group('I/O')
    groupIO.add_argument('mapfile', metavar='MapFile', type=str,
                         help='A QIIME-formatted mapping file with extra columns (see below)')
    groupIO.add_argument('--rows', type=str, default='all',
                         help='Which rows of the mapping file to use (eg., "all"=all rows; "1-48"=rows1-48; "1,3,5-6"=rows1+3+5+6), (default: %(default)s)')
    groupIO.add_argument('--prefix', type=str, default='TECAN_NGS_amplicon',
                         help='Output file name prefix (default: %(default)s)')

    ## Destination plate
    dest = parser.add_argument_group('Destination plate')
    dest.add_argument('--dest', type=str, default='96 Well[008]',
                      help='Distination labware ID on TECAN workbench (default: %(default)s)')
    dest.add_argument('--desttype', type=str, default='96',
                      choices=['96','384'],
                      help='Destination plate labware type (default: %(default)s)')
    dest.add_argument('--deststart', type=int, default=1,
                      help='Start well number on destination plate (default: %(default)s)')
    dest.add_argument('--rxns', type=int, default=3,
                      help='Number of replicate PCRs per sample (default: %(default)s)')

    ## MasterMix
    mm = parser.add_argument_group('Master mix')
    mm.add_argument('--mmtube', type=int, default=1,
                        help='MasterMix tube number (default: %(default)s)')
    mm.add_argument('--mmvolume', type=float, default=13.1,
                        help='MasterMix volume per PCR (default: %(default)s)')

    ## Primers
    primers = parser.add_argument_group('Primers')
    primers.add_argument('--fpvolume', type=float, default=1.0,
                        help='Forward primer volume per PCR (default: %(default)s)')
    primers.add_argument('--rpvolume', type=float, default=1.0,
                        help='Reverse primer volume per PCR (default: %(default)s)')
    primers.add_argument('--fptube', type=int, default=0,
                        help='Forward non-bacode primer tube number (0 = barcoded primer on a plate), (default: %(default)s)')
    primers.add_argument('--rptube', type=int, default=0,
                        help='Reverse non-bacode primer tube number (0 = barcoded primer on a plate), (default: %(default)s)')

    # Liquid classes
    liq = parser.add_argument_group('Liquid classes')
    liq.add_argument('--mm-liq', type=str, default='MasterMix Free Multi',
                      help='Mastermix liquid class (default: %(default)s)')
    liq.add_argument('--primer-liq', type=str, default='Water Contact Wet Single',
                      help='Primer liquid class (default: %(default)s)')
    liq.add_argument('--sample-liq', type=str, default='Water Contact Wet Single',
                      help='Sample liquid class (default: %(default)s)')
    liq.add_argument('--water-liq', type=str, default='Water Contact Wet Single',
                      help='Water liquid class (default: %(default)s)')

    ## Misc
    misc = parser.add_argument_group('Misc')
    misc.add_argument('--pcrvolume', type=float, default=25.0,
                        help='Total volume per PCR (default: %(default)s)')
    misc.add_argument('--errorperc', type=float, default=10.0,
                        help='Percent of extra total reagent volume to include (default: %(default)s)')

    # running test args
    if test_args:
        args = parser.parse_args(test_args)
        return args


def check_args(args):
    """Checking user input
    """
    # destination start
    if args.desttype == '96':
        destlimit = 96
    elif args.desttype == '384':
        destlimit = 384
    if args.deststart < 1 or args.deststart > destlimit:
        msg = 'Destination start well # must be in range: 1-{}'
        raise ValueError(msg.format(destlimit))
    # destination labware
   # args.destlabware = {x.split(':')[0]:x.split(':')[1] for x in args.destlabware.split(',')}
    # rows in mapping file
    args.rows = Utils.make_range(args.rows, set_zero_index=True)
    # tube number
    if args.mmtube < 1 or args.mmtube > 24:
        msg = '{} tube # must be in range: 1-24'
        raise ValueError(msg.format('MasterMix'))
    if args.fptube < 0 or args.fptube > 24:
        msg = '{} tube # must be in range: 1-24 (or 0 if no tube)'
        raise ValueError(msg.format('Forward primer'))
    if args.rptube < 0 or args.rptube > 24:
        msg = '{} tube # must be in range: 1-24 (or 0 if no tube)'
        raise ValueError(msg.format('Reverse primer'))
    # volumes
    ## mastermix
    if args.mmvolume > args.pcrvolume:
        msg = 'MasterMix volume > total PCR volume'
        raise ValueError(msg)
    if args.mmvolume / 2.0 > args.pcrvolume:
        msg = 'WARNING: MasterMix volume > half of PCR volume'
        print(msg, file=sys.stderr)
    ## primers?

def map2df(mapfile, row_select=None):
    """Loading a mapping file as a pandas dataframe
    mapfile: string; mapping file path
    row_select: select particular rows of table in map file
    """
    # load via pandas IO
    if mapfile.endswith('.txt') or mapfile.endswith('.csv'):
        df = pd.read_csv(mapfile, sep='\t')
    elif mapfile.endswith('.xls') or mapfile.endswith('.xlsx'):
        xls = pd.ExcelFile(mapfile)
        df = pd.read_excel(xls)
    else:
        raise ValueError('Mapping file not in usable format')

    # selecting particular rows
    if row_select is not None:
        df = df.iloc[row_select]

    # return
    return df

def check_df_map(df_map, args):
    """Assertions of df_map object formatting
    * Assumes `sample` field = 1st column
    df_map: map dataframe
    """
    # checking columns
    req_cols = ['TECAN_sample_labware', 'TECAN_sample_location',
                'TECAN_primer_labware', 'TECAN_primer_location',
                'TECAN_sample_rxn_volume']

    msg = 'Required column "{}" not found'
    for req_col in req_cols:
        if req_col not in df_map.columns.values:
            raise ValueError(msg.format(req_col))

    # checking for unique samples
    if any(df_map.duplicated(df_map.columns.values[0])):
        msg = 'WARNING: duplicated sample values in the mapping file'
        print(msg, file=sys.stderr)
    # checking for unique barcodes (NaN's filtered out)
    barcode_col = df_map.columns.values[1]
    df_tmp = df_map[df_map[barcode_col].notnull()]
    if any(df_tmp.duplicated(barcode_col)):
        msg = 'WARNING: duplicated barcodes in the mapping file'
        print(msg, file=sys.stderr)

    # checking sample volumes
    msg = 'WARNING: sample volume > mastermix volume'
    for sv in df_map['TECAN_sample_rxn_volume']:
        if sv > args.mmvolume:
            print(msg, file=sys.stderr)
        if sv < 0:
            raise ValueError('Sample volume < 0')

def add_dest(df_map, dest_labware, dest_type='96',
             dest_start=1, rxn_reps=3):
    """Setting destination locations for samples & primers.
    Making a new dataframe with:
      [sample, sample_rep, dest_labware, dest_location]
    * For each sample (i):
      * For each replicate (ii):
        * plate = destination plate type
        * well = i * (ii+1) + (ii+1) + start_offset
    Joining to df_map
    """
    dest_start= int(dest_start)

    # init destination df
    sample_col = df_map.columns[0]
    cols = [sample_col, 'TECAN_pcr_rxn_rep',
            'TECAN_dest_labware', 'TECAN_dest_location']
    ncol = len(cols)
    nrow = df_map.shape[0] * rxn_reps
    if dest_type == '96' and nrow > 97 - dest_start:
        nrow = 97 - dest_start
    elif dest_type == '384' and nrow > 385 - dest_start:
        nrow = 385 - dest_start
    df_dest = pd.DataFrame(np.nan, index=range(nrow), columns=cols)

    # filling destination df
    for i,(sample,rep) in enumerate(product(df_map.ix[:,0], range(rxn_reps))):
        # dest location
        dest_location = i + dest_start
        msg = 'WARNING: Not enough wells for the number of samples'
        msg = msg + '. Truncating to max samples that will fit on the plate'
        if dest_type == '96' and dest_location > 96:
            print(msg, file=sys.stderr)
            break
        elif dest_type == '384' and dest_location > 384:
            print(msg, file=sys.stderr)
            break
            # adding values DF
        df_dest.iloc[i] = [sample, rep+1, dest_labware, dest_location]

    # df join (map + destination)
    df_j = pd.merge(df_map, df_dest, on=sample_col, how='inner')
    assert df_j.shape[0] == df_dest.shape[0], 'map-dest DF join error'
    assert df_j.shape[0] > 0, 'DF has len=0 after adding destinations'

    # return
    return df_j


def reorder_384well(df, reorder_col):
    """Reorder values so that the odd, then the even locations are
    transferred. This is faster for a 384-well plate
    df: pandas.DataFrame
    reorder_col: column name to reorder
    """
    df['TECAN_sort_IS_EVEN'] = [x % 2 == 0 for x in df[reorder_col]]
    df.sort_values(by=['TECAN_sort_IS_EVEN', reorder_col], inplace=True)
    df = df.drop('TECAN_sort_IS_EVEN', 1)
    df.index = range(df.shape[0])
    return df


def pip_mastermix(df_map, outFH, mmvolume=13.1, mmtube=1, liq_cls='Water Free Single'):
    """Writing worklist commands for aliquoting mastermix.
    Using 1-asp-multi-disp with 200 ul tips.
    Method:
    * calc max multi-dispense for 200 ul tips & mmvolume
    * for 1:n_dispense
      * determine how many disp left for channel (every 8th)
      * if n_disp_left < n_disp: n_disp = n_disp_left
      * calc total volume: n_disp * mmvolume
    """
    mmtube = int(mmtube)
    outFH.write('C;MasterMix\n')

    MD = Fluent.multi_disp()
    MD.SrcRackLabel = 'micro15[{0:0>3}]'.format(mmtube)
    MD.SrcPosition = mmtube
    MD.DestRackLabel = df_map.ix[:,'TECAN_dest_labware']    
    MD.DestPositions = df_map.ix[:,'TECAN_dest_location']
    MD.Volume = mmvolume
    MD.LiquidClass = liq_cls
    MD.NoOfMultiDisp = int(np.floor(160 / mmvolume))  # using 200 ul tips

    outFH.write(MD.cmd() + '\n')


def pip_nonbarcode_primer(df_map, outFH, volume, tube, liq_cls='Water Free Single'):
    """Pipetting primers from tube.
    Assuming primer is aliquoted to all samples
    df_map : mapping file dataframe
    volume : volume to aliquot to each reaction
    tube : tube number (rackID assumed)
    liq_cls : liquid class
    """
    tube = int(tube)
    outFH.write('C;Non-barcoded primers\n')
    # for each Sample-PCR_rxn_rep, write out asp/dispense commands
    for i in range(df_map.shape[0]):
        # aspiration
        asp = Fluent.aspirate()
        asp.RackLabel = 'micro15[{0:0>3}]'.format(tube)
        asp.Position = tube
        asp.Volume = volume
        asp.LiquidClass = liq_cls
        outFH.write(asp.cmd() + '\n')

        # dispensing
        disp = Fluent.dispense()
        disp.RackLabel = df_map.ix[i,'TECAN_dest_labware']
        disp.Position = df_map.ix[i,'TECAN_dest_location']
        disp.Volume = volume
        disp.LiquidClass = liq_cls
        outFH.write(disp.cmd() + '\n')

        # tip to waste
        outFH.write('W;\n')

def pip_primers(df_map, outFH, fp_volume=0, rp_volume=0,
                fp_tube=0, rp_tube=0, liq_cls='Water Free Single'):
    """Commands for aliquoting primers
    """
    outFH.write('C;Primers\n')
    primer_plate_volume = 0
    # pipetting non-barcoded primers
    ## forward primer
    if fp_tube > 0 and fp_volume > 0:
        pip_nonbarcode_primer(df_map, outFH, fp_volume, fp_tube)
    else:
        primer_plate_volume += fp_volume
    ## reverse primer
    if rp_tube > 0 and rp_volume > 0:
        pip_nonbarcode_primer(df_map, outFH, rp_volume, rp_tube)
    else:
        primer_plate_volume += rp_volume

    # pipetting barcoded primers
    outFH.write('C;Barcoded primers\n')
    ## for each Sample-PCR_rxn_rep, write out asp/dispense commands
    for i in range(df_map.shape[0]):
        # aspiration
        asp = Fluent.aspirate()
        asp.RackLabel = df_map.ix[i,'TECAN_primer_labware']
        asp.Position = df_map.ix[i,'TECAN_primer_location']
        asp.Volume = primer_plate_volume
        asp.LiquidClass = liq_cls
        outFH.write(asp.cmd() + '\n')

        # dispensing
        disp = Fluent.dispense()
        disp.RackLabel = df_map.ix[i,'TECAN_dest_labware']
        disp.Position = df_map.ix[i,'TECAN_dest_location']
        disp.Volume = primer_plate_volume
        asp.LiquidClass = liq_cls
        outFH.write(disp.cmd() + '\n')

        # tip to waste
        outFH.write('W;\n')

def pip_samples(df_map, outFH, liq_cls='Water Free Single'):
    """Commands for aliquoting samples to each PCR rxn
    """
    outFH.write('C;Samples\n')
    # for each Sample-PCR_rxn_rep, write out asp/dispense commands
    for i in range(df_map.shape[0]):
        # aspiration
        asp = Fluent.aspirate()
        asp.RackLabel = df_map.ix[i,'TECAN_sample_labware']
        asp.Position = df_map.ix[i,'TECAN_sample_location']
        asp.Volume = df_map.ix[i,'TECAN_sample_rxn_volume']
        asp.LiquidClass = liq_cls
        outFH.write(asp.cmd() + '\n')

        # dispensing
        disp = Fluent.dispense()
        disp.RackLabel = df_map.ix[i,'TECAN_dest_labware']
        disp.Position = df_map.ix[i,'TECAN_dest_location']
        disp.Volume = df_map.ix[i,'TECAN_sample_rxn_volume']
        disp.LiquidClass = liq_cls
        outFH.write(disp.cmd() + '\n')

        # tip to waste
        outFH.write('W;\n')

def pip_water(df_map, outFH, pcr_volume=25.0, mm_volume=13.1, 
              fp_volume=2.0, rp_volume=2.0, liq_cls='Water Free Single'):
    """Commands for aliquoting water to each PCR rxn
    """
    outFH.write('C;Water\n')
    # calculate the amount of water
    water_volume = []
    for i in range(df_map.shape[0]):
        samp_volume = df_map.ix[i,'TECAN_sample_rxn_volume']
        w_need = pcr_volume - (samp_volume + mm_volume + fp_volume + rp_volume)
        assert w_need >= 0, 'Water volume is negative: {}'.format(w_need)
        water_volume.append(w_need)
    df_map['TECAN_water_rxn_volume'] = water_volume

    # for each Sample-PCR_rxn_rep, write out asp/dispense commands
    for i in range(df_map.shape[0]):
        # aspiration
        asp = Fluent.aspirate()
        asp.RackLabel = df_map.ix[i,'TECAN_sample_labware']
        asp.Position = df_map.ix[i,'TECAN_sample_location']
        asp.Volume = water_volume[i]
        asp.LiquidClass = liq_cls
        outFH.write(asp.cmd() + '\n')

        # dispensing
        disp = Fluent.dispense()
        disp.RackLabel = df_map.ix[i,'TECAN_dest_labware']
        disp.Position = df_map.ix[i,'TECAN_dest_location']
        disp.LiquidClass = liq_cls
        disp.Volume = water_volume[i]
        outFH.write(disp.cmd() + '\n')

        # tip to waste
        outFH.write('W;\n')

def add_error(x, error_perc):
    if x is None:
        return None
    return x * (1.0 + error_perc / 100.0)

def write_report_line(outFH, subject, volume, round_digits=1, error_perc=None):
    if volume is None:
        v = 'NA'
    else:
        if error_perc is not None:
            volume = add_error(volume, error_perc)
        v = round(volume, round_digits)
    outFH.write('{}:\t{}\n'.format(subject, v))

def write_report(df_map, outFH, pcr_volume, mm_volume,
                 fp_tube, fp_volume, rp_tube, rp_volume,
                 n_rxn_reps, error_perc=10.0):
    """Writing a report on
    """
    # calculating total volumes
    n_rxn = df_map.shape[0]
    ## total PCR
    total_pcr_volume = pcr_volume * n_rxn
    ## total mastermix
    total_mm_volume = pcr_volume * n_rxn
    ## total primer
    if fp_tube > 0 and fp_volume > 0:
        total_fp_volume = fp_volume * n_rxn
    else:
        total_fp_volume = None
    if rp_tube > 0 and rp_volume > 0:
        total_rp_volume = rp_volume * n_rxn
    else:
        total_rp_volume = None
    ## total water
    total_water_volume = sum(df_map['TECAN_water_rxn_volume'])

    # report
    # number of samples
    outFH.write('# PCR REPORT\n')
    outFH.write('Number of PCR replicates:\t{}\n'.format(n_rxn_reps))
    outFH.write('Number of total PCRs:\t{}\n'.format(n_rxn))
    ## rxn volumes
    outFH.write('# Volumes per PCR (ul)\n')
    write_report_line(outFH, 'Total', pcr_volume)
    write_report_line(outFH, 'Master Mix', mm_volume)
    write_report_line(outFH, 'Forward Primer', fp_volume)
    write_report_line(outFH, 'Reverse Primer', rp_volume)
    ## raw total volumes
    outFH.write('# Total volumes (ul)\n')
    write_report_line(outFH, 'Master Mix', total_mm_volume)
    write_report_line(outFH, 'Forward Primer', total_fp_volume)
    write_report_line(outFH, 'Reverse Primer', total_rp_volume)
    write_report_line(outFH, 'Water', total_water_volume)
    ## total volumes with error
    outFH.write('# Total volumes with (ul; {}% error)\n'.format(error_perc))
    write_report_line(outFH, 'Master Mix', total_mm_volume, error_perc=error_perc)
    write_report_line(outFH, 'Forward Primer', total_fp_volume, error_perc=error_perc)
    write_report_line(outFH, 'Reverse Primer', total_rp_volume, error_perc=error_perc)
    write_report_line(outFH, 'Water', total_water_volume, error_perc=error_perc)
    # samples
    outFH.write('')


def main(args=None):
    # Input
    if args is None:
        args = parse_args()
    check_args(args)
    # Import
    df_map = map2df(args.mapfile, row_select=args.rows)
    check_df_map(df_map, args)
    # Making destination dataframe
    df_map = add_dest(df_map, args.dest,
                      dest_type=args.desttype,
                      dest_start=args.deststart,
                      rxn_reps=args.rxns)

    # Reordering dest if plate type is 384-well
    if args.desttype == '384':
        df_map = reorder_384well(df_map, 'TECAN_dest_location')
    elif args.desttype == '96':
        pass
    else:
        msg = 'Destination labware type "{}" not recognized'
        msg.format(args.desttype)

    # GWL file construction
    ## gwl open
    gwl_file = args.prefix + '.gwl'
    with open(gwl_file, 'w') as gwlFH:
        ## mastermix
        pip_mastermix(df_map, gwlFH,
                      mmtube=args.mmtube,
                      mmvolume=args.mmvolume, 
                      liq_cls=args.mm_liq)
        ## primers
        pip_primers(df_map, gwlFH,
                    fp_volume=args.fpvolume,
                    rp_volume=args.rpvolume,
                    fp_tube=args.fptube,
                    rp_tube=args.rptube, 
                    liq_cls=args.primer_liq)
        ## samples
        pip_samples(df_map, gwlFH, 
                    liq_cls=args.sample_liq)
        ## water
        pip_water(df_map, gwlFH,
                  pcr_volume=args.pcrvolume,
                  mm_volume=args.mmvolume,
                  fp_volume=args.fpvolume,
                  rp_volume=args.rpvolume, 
                  liq_cls=args.water_liq)

    # Report (total volumes; sample truncation; samples)
    report_file = args.prefix + '.report'
    with open(report_file, 'w') as repFH:
        write_report(df_map, outFH=repFH,
                     pcr_volume=args.pcrvolume,
                     mm_volume=args.mmvolume,
                     fp_tube=args.fptube,
                     fp_volume=args.fpvolume,
                     rp_tube=args.rptube,
                     rp_volume=args.rpvolume,
                     n_rxn_reps=args.rxns,
                     error_perc=args.errorperc)

    # Mapping file with destinations
    df_file = args.prefix + '_map.txt'
    df_map['TECAN_water_rxn_volume'] = df_map['TECAN_water_rxn_volume'].round(2)
    df_map['TECAN_dest_location'] = df_map['TECAN_dest_location'].astype(int)
    df_map['TECAN_pcr_rxn_rep'] = df_map['TECAN_pcr_rxn_rep'].astype(int)
    df_map.to_csv(df_file, sep='\t', index=False, na_rep='NA')

    # Create windows-line breaks formatted versions
    gwl_file_win = Utils.to_win(gwl_file)
    report_file_win = Utils.to_win(report_file)
    df_file_win = Utils.to_win(df_file)

    # Return
    return (gwl_file, gwl_file_win,
            report_file, report_file_win, 
            df_file, df_file_win)

# main
if __name__ == '__main__':
    pass





# -*- coding: utf-8 -*-
# import
## batteries
from __future__ import print_function
import os
import sys
import argparse
from itertools import product
import string
## 3rd party
import numpy as np
import pandas as pd
## package
from leylab_pipelines import Utils
from leylab_pipelines import Fluent


# functions
def get_desc():
    desc = 'Create robot commands for qPCR setup'
    return desc

def parse_args(test_args=None, subparsers=None):
    # desc
    desc = get_desc()
    epi = """DESCRIPTION:
    Create a worklist file for the TECAN Fluent robot for qPCR setup.
    The input is an exported plate layout from the BioRad qPCR software.
    The file format should be Excel or CSV.
    Just create a plate layout for your experimet, then export and add some needed info:
      The following columns should also be added to the table:
      * "Sample labware"  (labware containing sample DNA/RNA; eg., "96-Well[001]")
      * "Sample location"  (numeric; minimum of 1)
      * "Sample volume"  (numeric)
      * "MM name"  (Name of master mix; this allows for multiple mastermixes)
      * "MM volume"  (Volume of master mix in PCR rxn)
      * "Water volume"  (Volume of water in PCR rxn)
    
    Notes:
    * Sample locations in plates numbered are column-wise. 
    * The setup file (input table) MUST have a header (capitalization doesn't matter)
    * All volumes are in ul.
    """
    if subparsers:
        parser = subparsers.add_parser('qPCR', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # args
    ## I/O
    groupIO = parser.add_argument_group('I/O')
    groupIO.add_argument('setup', metavar='SetupFile', type=str,
                         help='An Excel or CSV file with experimental setup')
    groupIO.add_argument('--prefix', type=str, default='TECAN_qPCR',
                         help='Output file name prefix (default: %(default)s)')
    groupIO.add_argument('--format', type=str, default=None,
                        help='File format (Excel or CSV). If not provided, the format is determined from the file extension (default: %(default)s)') 

    ## destination plate
    dest = parser.add_argument_group('Destination labware')
    dest.add_argument('--dest', type=str, default='384 Well[004]',
                      help='Destination plate labware ID on TECAN worktable (default: %(default)s)')
    dest.add_argument('--desttype', type=str, default='384',
                      choices=['96', '384'],
                      help='Destination plate labware type (default: %(default)s)')

    ## source labware 
    src = parser.add_argument_group('Source labware')
    src.add_argument('--mm', type=str, default='Tubes[001]',
                      help='Mastermix source labware ID on TECAN worktable (default: %(default)s)')
    src.add_argument('--mmloc', type=int, default=1,
                      help='Mastermix start position on source labware (default: %(default)s)')
    src.add_argument('--water', type=str, default='100ml[001]',
                      help='Water source labware ID on TECAN worktable (default: %(default)s)')
    src.add_argument('--waterloc', type=int, default=1,
                      help='Water start position on source labware (default: %(default)s)')

    # liquid classes
    liq = parser.add_argument_group('Liquid classes')
    liq.add_argument('--mmliq', type=str, default='MasterMix Free Multi',
                      help='Mastermix liquid class (default: %(default)s)')
    liq.add_argument('--sampliq', type=str, default='Water Contact Wet Single',
                      help='Sample liquid class (default: %(default)s)')
    liq.add_argument('--waterliq', type=str, default='Water Contact Wet Single',
                      help='Water liquid class (default: %(default)s)')

    # parse & return
    if test_args:
        args = parser.parse_args(test_args)
        return args


def check_args(args):
    """Checking user input
    """
    pass


def load_setup(input_file, file_format=None, header=0):
    # format
    if file_format is None:
        if input_file.endswith('.txt'):
            file_format = 'csv'
        elif input_file.endswith('.csv'):
            file_format = 'txt'
        elif input_file.endswith('.xls') or input_file.endswith('.xlsx'):
            file_format = 'excel'
    else:
        file_format = file_format.lower()
        
    # load via pandas IO
    if file_format == 'csv':
        df = pd.read_csv(input_file, sep=';', header=header)
    elif file_format == 'tab':
        df = pd.read_csv(input_file, sep='\t', header=header)
    elif file_format == 'excel':
        xls = pd.ExcelFile(input_file)
        df = pd.read_excel(xls, header=header)
    else:
        raise ValueError('Setup file not in usable format')

    # standarizing column IDs
    df.columns = [x.lower() for x in df.columns]

    # assert & return
    assert df.shape[1] > 1, 'Input file is only 1 column; wrong delimiter used?'    
    return df


def check_df_setup(df_setup):
    """Assertions of df_conc object formatting
    """
    # checking for column IDs
    col_IDs = ('row', 'column', 'sample type', 
               'sample labware', 'sample location', 'sample volume',
               'mm name', 'mm volume', 'water volume')
    msg = 'Column "{}" not found (captilization invariant)'
    for x in col_IDs:
        if not x in df_setup.columns:
            raise ValueError(msg.format(x))

    # checking sample locations (>=1)
    msg = 'ERROR (SetupFile, line={}): location is < 1'
    for i,loc in enumerate(df_setup['sample location']):
        if loc < 1:
            print(msg.format(i), file=sys.stderr)
    
    # checking sample conc
    msg = 'ERROR (setupFile, line={}): volume is < 0'
    for i,vol in enumerate(df_setup['sample volume']):
        assert np.isnan(vol) or vol >= 0.0, msg.format(i)
    for i,vol in enumerate(df_setup['mm volume']):
        assert np.isnan(vol) or vol >= 0.0, msg.format(i)
    for i,vol in enumerate(df_setup['water volume']):
        assert np.isnan(vol) or vol >= 0.0, msg.format(i)
    

def plate2robot_loc(row_val, col_val, plate_type='96'):
    """Changing positioning from row (letter) and column (number)
    to just numeric position (column-wise) on the plate,
    which is needed for TECAN robot positioning. 
    Using index for identifying well number in plate
    [args]
    row_val: string
    col_vol: string
    plate_type: string; plate type to determine well location indexing
    """    
    # index for converting row to numeric
    idx = string.ascii_uppercase
    idx = {x:i+1 for i,x in enumerate(idx)}    
    row_val = idx[row_val]

    # getting location on plate
    msg = 'Destination location "{}" is out of range'
    if plate_type == '96':
        loc = (col_val - 1) * 8 + row_val
        assert loc > 0 and loc <= 96, msg.format(loc)
    elif plate_type == '384':
        loc = (col_val - 1) * 16 + row_val
        assert loc > 0 and loc <= 384, msg.format(loc)
    else:
        msg = 'Labware type "{}" not recognized'
        raise ValueError(msg.format(plate_type))
    return loc


def add_dest(df_setup, dest_labware, dest_type='96'):
    """Setting destination locations for samples & reagents
    Adding to df_conc:
      [dest_labware, dest_location]
    """
    # setting destination labware
    df_setup['dest_labware'] = dest_labware
    
    # setting destination location based on plate layout 
    func = lambda x: plate2robot_loc(x['row'], x['column'], plate_type=dest_type)
    df_setup['dest_location'] = df_setup.apply(func, 1)


def reorder_384well(df, reorder_col):
    """Reorder values so that the odd, then the even locations are
    transferred. This is faster for a 384-well plate
    df: pandas.DataFrame
    reorder_col: column name to reorder
    """
    df['TECAN_sort_IS_EVEN'] = [x % 2 == 0 for x in df[reorder_col]]
    df.sort_values(by=['TECAN_sort_IS_EVEN', reorder_col], inplace=True)
    df = df.drop('TECAN_sort_IS_EVEN', 1)
    df.index = range(df.shape[0])
    return df


def pip_mastermix(df_setup, outFH, src_labware, src_start=1, liq_cls='Water Free Single'):
    """Writing worklist commands for aliquoting mastermix.
    Using 1-asp-multi-disp with certain tips.
    """
    # split by mastermix names (different master mixes)
    MM_names = np.unique(df_setup['mm name'])

    outFH.write('C;Master mix\n')    
    for i,name in enumerate(MM_names):
        df = df_setup.loc[df_setup['mm name'] == name]
        # filtering out 'nan' values
        df = df.loc[pd.notnull(df_setup['mm volume'])]
        df.index = range(df.shape[0])
        
        # determing how many multi-disp
        max_vol = max(df['mm volume'])
        if max_vol > 160.0:
            n_disp = int(np.floor(900.0 / max_vol))  # using 1000 ul tips
        else:
            n_disp= int(np.floor(160.0 / max_vol))   # using 200 ul tips

        # making multi-disp object
        MD = Fluent.multi_disp()
        MD.SrcRackLabel = src_labware
        MD.SrcPosition = src_start + i
        MD.DestRackLabel = df.dest_labware
        MD.DestPositions = df.dest_location
        MD.Volume = df['mm volume']
        MD.NoOfMultiDisp = n_disp
        MD.LiquidClass = liq_cls
        # writing
        outFH.write(MD.cmd() + '\n')


def pip_samples(df_setup, outFH, liq_cls='Water Contact Wet Single'):
    """Commands for aliquoting samples into distination plate
    """
    outFH.write('C;Samples\n')
    # filtering 'nan' from table
    x = pd.notnull(df_setup['sample labware'])
    y = pd.notnull(df_setup['sample location'])
    df = df_setup.loc[x & y]
    df.index = range(df.shape[0])
    
    # for each Sample, write out asp/dispense commands
    for i in range(df.shape[0]):
        # aspiration
        asp = Fluent.aspirate()
        asp.RackLabel = df.ix[i,'sample labware']
        asp.Position = df.ix[i,'sample location']
        asp.Volume = round(df.ix[i,'sample volume'], 2)
        asp.LiquidClass = liq_cls
        outFH.write(asp.cmd() + '\n')

        # dispensing
        disp = Fluent.dispense()
        disp.RackLabel = df.ix[i,'dest_labware']
        disp.Position = df.ix[i,'dest_location']
        disp.Volume = round(df.ix[i,'sample volume'], 2)
        disp.LiquidClass = liq_cls
        outFH.write(disp.cmd() + '\n')

        # tip to waste
        outFH.write('W;\n')

def pip_water(df_setup, outFH, src_labware, src_start=1, 
              liq_cls='Water Contact Wet Single'):
    """Writing worklist commands for aliquoting water
    Using single asp-disp.
    """

    # filtering 'nan' from table
    x = pd.notnull(df_setup['water volume'])
    df = df_setup.loc[x]
    df.index = range(df.shape[0])
            
    for i in range(df.shape[0]):
        # aspiration
        asp = Fluent.aspirate()
        asp.RackLabel = src_labware
        asp.Position = src_start
        asp.Volume = round(df.ix[i,'water volume'], 2)
        asp.LiquidClass = liq_cls
        outFH.write(asp.cmd() + '\n')

        # dispensing
        disp = Fluent.dispense()
        disp.RackLabel = df.ix[i,'dest_labware']
        disp.Position = df.ix[i,'dest_location']
        disp.Volume = round(df.ix[i,'water volume'], 2)
        disp.LiquidClass = liq_cls
        outFH.write(disp.cmd() + '\n')

        # tip to waste
        outFH.write('W;\n')

def add_error(x, error_perc):
    if x is None:
        return None
    return x * (1.0 + error_perc / 100.0)

def write_report_line(outFH, subject, volume, round_digits=1, error_perc=None):
    if volume is None:
        v = 'NA'
    else:
        if error_perc is not None:
            volume = add_error(volume, error_perc)
        v = round(volume, round_digits)
    outFH.write('{}:\t{}\n'.format(subject, v))

def write_report(df_setup, outFH):
    """Writing a report on the qPCR setup
    """
    # calculating total volumes
    n_rxn = df_setup.shape[0]
    ## total mastermix
    total_mm_volume = np.sum(df_setup['mm volume'])
    ## total water
    total_water_volume = np.sum(df_setup['water volume'])

    # report
    # number of samples
    outFH.write('# PCR REPORT\n')
    outFH.write('Number of total rxns:\t{}\n'.format(n_rxn))
    ## raw total volumes
    outFH.write('# Total reagent volumes (ul)\n')
    write_report_line(outFH, 'Master Mix', total_mm_volume)
    write_report_line(outFH, 'Water', total_water_volume)
    ## with pipetting error
    outFH.write('# Total reagent volumes + 10% more (ul)\n')
    write_report_line(outFH, 'Master Mix', total_mm_volume, error_perc=10)
    write_report_line(outFH, 'Water', total_water_volume, error_perc=10)
    
    # samples
    outFH.write('')

    
def main(args=None):
    # Input
    if args is None:
        args = parse_args()
#    check_args(args)

    # Load input table
    df_setup = load_setup(args.setup, 
                          file_format=args.format)
    check_df_setup(df_setup)

    # adding sample/reagent destinations to setup table
    add_dest(df_setup, args.dest, dest_type=args.desttype)
        
    # Reordering dest if plate type is 384-well
    if args.desttype == '384':
         df_setup = reorder_384well(df_setup, 'dest_location')
    elif args.desttype == '96':
        pass
    else:
        msg = 'Labware type "{}" not recognized'
        raise ValueError(msg.format(args.desttype))
    
    # Writing out gwl file
    gwl_file = args.prefix + '.gwl'
    with open(gwl_file, 'w') as gwlFH:
        ## Master mix(es)
        pip_mastermix(df_setup, outFH=gwlFH, 
                      src_labware=args.mm,
                      src_start=args.mmloc, 
                      liq_cls=args.mmliq)
        ## Samples
        pip_samples(df_setup, outFH=gwlFH, liq_cls=args.sampliq)
        ## Water
        pip_water(df_setup, outFH=gwlFH, src_labware=args.water, 
                  src_start=args.waterloc, liq_cls=args.waterliq)

    # Creating report file
    report_file = args.prefix + '_report.txt'
    with open(report_file, 'w') as repFH:
        write_report(df_setup, outFH=repFH)
        

    # Create windows-line breaks formatted versions
    gwl_file_win = Utils.to_win(gwl_file)
    report_file_win = Utils.to_win(report_file)

    # end
    return(gwl_file, gwl_file_win, report_file, report_file_win)


# main
if __name__ == '__main__':
    pass





#!/usr/bin/env python

from __future__ import print_function

import os
import sys
from leylab_pipelines import Dilute



# main
if __name__ == '__main__':
    files = Dilute.main()
    for F in files:
        print('File written: {}'.format(F), file=sys.stderr)



# import
## batteries
import argparse
import requests
import sys
import re
import warnings
import multiprocessing as mp
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
## 3rd party
import xmltodict


# functions
def get_desc():
    desc = 'Convert between various gene or protein IDs'
    return desc

def parse_args(test_args=None, subparsers=None):
    # desc
    desc = get_desc()
    epi = """DESCRIPTION:
    Convert between various gene or protein IDs using entrez API requests.

    Entrez requests can be done in parallel. 

    IDs:
      If the IDs are provided in a table, select the column delimiter and 
      column number (1-indexed). 
      The list of IDs can be provided via STDIN by using `STDIN`
    """
    if subparsers:
        parser = subparsers.add_parser('convert', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # args
    io = parser.add_argument_group('Input/Output')
    io.add_argument('IDs', metavar='IDs', type=str,
                     help='A file containing IDs. See description.')
    io.add_argument('-m', '--method', default='Accession2Taxonomy',
                    choices=['Accession2Taxonomy', 
                             'EntrezGene2Uniprot',
                             'Uniprot2EntrezGene',
                             'HGNC2EntrezGene',
                             'EntrezTranGene2EntrezGene'],
                    help='ID conversion method. (default: %(default)s)')
    io.add_argument('-s', '--sep', default='\t',
                    help='Column separator (default: %(default)s)')
    io.add_argument('-c', '--column', type=int, default=1,
                    help='Column containing the IDs (default: %(default)s)')
    io.add_argument('-o', '--outfile', default='-',
                     help='Output file name; "-" if to STDOUT (default: %(default)s)')
    io.add_argument('-e', '--email', default='dummyemail@dummybunny.info',
                     help='user email address (default: %(default)s)')

    misc = parser.add_argument_group('Misc')
    misc.add_argument('-p', '--procs', type=int, default=1,
                      help='Number of processors to use (default: %(default)s)')

    # running test args
    if test_args:
        args = parser.parse_args(test_args)
        return args


def main(args=None):
    # Input
    if args is None:
        args = parse_args()

    # reading IDs
    IDs = read_IDs(args.IDs, sep=args.sep, column=args.column)

    # conversion    
    IDs = get_conversion(IDs, args.email, args.method, procs=args.procs)
    
    # writing IDs
    write_IDs(IDs, args.outfile)


def write_IDs(IDs, outfile='-'):
    """writing out IDs
    """
    if outfile == '-':
        outF = sys.stdout
    else:
        outF = open(outfile, 'w')

    header = '\t'.join(['orig_ID', 'new_ID'])
    outF.write(header + '\n')
    for x in IDs:
        outF.write('\t'.join(x) + '\n')

    if outfile != '-':
        outF.close()    
    

def get_conversion(IDs, email, method, procs=1):
    conv = Conversion(email)

    if procs < 2:
        procs = None
    pool = mp.Pool(processes = procs)

    if method.lower() == 'accession2taxonomy':
        IDs = pool.map(conv.convert_accession_to_taxid, IDs)
    elif method.lower() == 'entrezgene2uniprot':
        IDs = pool.map(conv.convert_entrez_to_uniprot, IDs)
    elif method.lower() == 'uniprot2entrezgene':
        IDs = pool.map(convert_uniprot_to_entrez, IDs)
    elif method.lower() == 'hgnc2entrezgene':
        IDs = pool.map(convert_hgnc_to_entrez, IDs)
    elif method.lower() == 'entreztrangene2entrezgene':
        IDs = poolmap(convert_ensembl_to_entrez, IDs)
    else:
        msg = 'Method "{}" not recognized'
        raise IOError(msg.format(method))
    return IDs
    

def read_IDs(infile, sep='\t', column=1):
    """Reading in IDs
    """
    column = column - 1
    if infile == 'STDIN':
        inF = sys.stdin
    else:
        inF = open(infile, 'r')

    IDs = []
    for line in inF:
        line = line.rstrip().split(sep)
        IDs.append(line[column])

    if infile != 'STDIN':
        inF.close()
    return IDs


class Conversion(object):
    def __init__(self, email):
        """email is required
        """
        self.params = {}
        self.email = email
        self.params['tool'] = 'PyEntrez'
        if re.match(r"[^@]+@[^@]+\.[^@]+", self.email):
            pass
        else:
            raise ValueError("Enter a valid Email Address")
        self.params["email"] = email
        self.options = urlencode(self.params, doseq=True)
        
    def convert_ensembl_to_entrez(self, ensembl):
        """Convert Ensembl Id to Entrez Gene Id
        """        
        if 'ENST' in ensembl:
                pass
        else:
            raise(IndexError)
        # Submit resquest to NCBI eutils/Gene database
        server = "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?" + self.options + "&db=gene&term={0}".format(ensembl)
        r = requests.get(server, headers={"Content-Type": "text/xml"})
        if not r.ok:
            r.raise_for_status()
            sys.exit()
        # Process Request
        response = r.text
        info = xmltodict.parse(response)
        try:
            geneId = info['eSearchResult']['IdList']['Id']
        except TypeError:
            sys.stderr.write('WARNING: No Entrez ID for "{}"\n'.format(ensembl))
            geneId = None
        return [ensembl, geneId]

    def convert_hgnc_to_entrez(self, hgnc):
        """Convert HGNC Id to Entrez Gene Id
        """
        entrezdict = {}
        server = "http://rest.genenames.org/fetch/hgnc_id/{0}".format(hgnc)
        r = requests.get(server, headers={ "Content-Type" : "application/json"})
        if not r.ok:
            r.raise_for_status()
            sys.exit()
        response = r.text
        info = xmltodict.parse(response)
        try:
            for data in info['response']['result']['doc']['str']:
                if data['@name'] == 'entrez_id':
                    entrezdict[data['@name']] = data['#text']
                if data['@name'] == 'symbol':
                    entrezdict[data['@name']] = data['#text']
        except KeyError:
            sys.stderr.write('WARNING: No Entrez ID for "{}"\n'.format(hgnc))            
            entrezdict = None
        try:
            entrezdict = entrezdict['entrez_id']
        except (KeyError, TypeError):
            entrezdict = None

        return [hgnc, entrezdict]

    def convert_entrez_to_uniprot(self, entrez):
        """Convert Entrez Id to Uniprot Id
        """
        server = "http://www.uniprot.org/uniprot/?query=%22GENEID+{0}%22&format=xml".format(entrez)
        r = requests.get(server, headers={ "Content-Type" : "text/xml"})
        if not r.ok:
            r.raise_for_status()
            sys.exit()
        response = r.text
        info = xmltodict.parse(response)
        try:
            data = info['uniprot']['entry']['accession'][0]
        except TypeError:
            data = info['uniprot']['entry'][0]['accession'][0]
        return [entrez, data]

    def convert_uniprot_to_entrez(self, uniprot):
        """Convert Uniprot Id to Entrez Id
        """
        # Submit request to NCBI eutils/Gene Database
        server = "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?" + self.options + "&db=gene&term={0}".format(uniprot)
        r = requests.get(server, headers={ "Content-Type" : "text/xml"})
        if not r.ok:
            r.raise_for_status()
            sys.exit()
        # Process Request
        response = r.text
        info = xmltodict.parse(response)
        geneId = info['eSearchResult']['IdList']['Id']
        # check to see if more than one result is returned
        # if you have more than more result then check which Entrez Id returns the same uniprot Id entered.
        if len(geneId) > 1:
            for x in geneId:
                c = self.convert_entrez_to_uniprot(x)[1]
                c = c.lower()
                u = uniprot.lower()
                if c==u:
                    return [uniprot, x]
        else:
            return [uniprot, geneId]

    def convert_accession_to_taxid(self, accessionid):
        """Convert Accession Id to Tax Id
        """
        # Submit request to NCBI eutils/Taxonomy Database
        server = "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?" + self.options + "&db=nuccore&id={0}&retmode=xml".format(accessionid)
        r = requests.get(server, headers={ "Content-Type" : "text/xml"})
        if not r.ok:
            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError:
                sys.stderr.write('WARNING: No taxonomy ID for "{}"\n'.format(accessionid)) 
            return [accessionid, None]
        # Process Request
        response = r.text
        records = xmltodict.parse(response)
        try:
            for i in records['GBSet']['GBSeq']['GBSeq_feature-table']['GBFeature']['GBFeature_quals']['GBQualifier']:
                for key, value in i.items():
                    if value == 'db_xref':
                        taxid = i['GBQualifier_value']
                        taxid = taxid.split(':')[1]
                        return [accessionid, taxid]
        except:
            for i in records['GBSet']['GBSeq']['GBSeq_feature-table']['GBFeature'][0]['GBFeature_quals']['GBQualifier']:
                for key, value in i.items():
                    if value == 'db_xref':
                        taxid = i['GBQualifier_value']
                        taxid = taxid.split(':')[1]
                        return [accessionid, taxid]
        return [accessionid, None]



# -*- coding: utf-8 -*-
# import
## batteries
from __future__ import print_function
import os
import sys
import argparse
from itertools import product
## 3rd party
import numpy as np
import pandas as pd
## package
from leylab_pipelines import Utils
from leylab_pipelines import Fluent


# functions
def get_desc():
    desc = 'Create robot commands for diluting samples'
    return desc

def parse_args(test_args=None, subparsers=None):
    # desc
    desc = get_desc()
    epi = """DESCRIPTION:
    Create a worklist file for the TECAN Fluent robot for diluting samples.
    The input is an Excel or tab-delimited file with:
    * Sample labware  (eg., "96 Well[001]")
    * Sample location (numeric value; minimum of 1)
    * Sample concentration (numeric value; units=ng/ul)
    
    Notes:
    * You can designate the input table columns for each value (see options).
    * Sample locations in plates numbered are column-wise. 
    * All volumes are in ul.
    """
    if subparsers:
        parser = subparsers.add_parser('dilute', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # args
    ## I/O
    groupIO = parser.add_argument_group('I/O')
    groupIO.add_argument('concfile', metavar='ConcFile', type=str,
                         help='An excel or tab-delim file of concentrations')
    groupIO.add_argument('--prefix', type=str, default='TECAN_dilute',
                         help='Output file name prefix (default: %(default)s)')

    ## concentration file
    conc = parser.add_argument_group('Concentation file')
    conc.add_argument('--format', type=str, default=None,
                        help='File format (excel or tab). If not provided, the format is determined from the file extension') 
    conc.add_argument('--header', action='store_false', default=True,
                        help='Header in the file? (default: %(default)s)')
    conc.add_argument('--labware', type=int, default=1,
                        help='Column containing the sample labware IDs (default: %(default)s)') 
    conc.add_argument('--location', type=int, default=2,
                        help='Column containing the sample location numbers (default: %(default)s)')
    conc.add_argument('--conc', type=int, default=3,
                        help='Column containing the sample concentrations (default: %(default)s)')
    conc.add_argument('--rows', type=str, default='all',
                      help='Which rows (not including header) of the column file to use ("all"=all rows; "1-48"=rows 1-48) (default: %(default)s)')

    ## dilution
    dil = parser.add_argument_group('Dilution')
    dil.add_argument('--dilution', type=float, default=1.0,
                     help='Target dilution concentration (ng/ul) (default: %(default)s)')
    dil.add_argument('--minvolume', type=float, default=2.0,
                     help='Minimum sample volume to use (default: %(default)s)')
    dil.add_argument('--maxvolume', type=float, default=30.0,
                     help='Maximum sample volume to use (default: %(default)s)')
    dil.add_argument('--mintotal', type=float, default=10.0,
                     help='Minimum post-dilution total volume (default: %(default)s)')
    dil.add_argument('--dlabware', type=str, default='100ml[001]',
                     help='Labware containing the dilutant (default: %(default)s)')

    ## destination plate
    dest = parser.add_argument_group('Destination plate')
    dest.add_argument('--dest', type=str, default='96 Well[001]',
                      help='Destination plate labware ID on TECAN worktable (default: %(default)s)')
    dest.add_argument('--desttype', type=str, default='96',
                      choices=['96','384'],
                      help='Destination plate labware type (default: %(default)s)')
    dest.add_argument('--deststart', type=int, default=1,
                      help='Start well number on destination OD plate (default: %(default)s)')
#    dest.add_argument('--destlabware', type=str, default='384 Well[004]',
#                      help='Destination Labware ID on the TECAN worktable (default: %(default)s)')
#                      default='96-well:96 Well[008],384-well:384 Well[004]',
#                      help='Choices for the destination labware name base on --desttype')

    # parse & return
    if test_args:
        args = parser.parse_args(test_args)
        return args


def check_args(args):
    """Checking user input
    """
    # input table column IDs
    assert args.labware >= 1, '--labware must be >= 1'
    assert args.location >= 1, '--location must be >= 1'
    assert args.conc >= 1, '--conc must be >= 1'
    # input table row select
    args.rows = Utils.make_range(args.rows, set_zero_index=True)
    # dilution
    assert args.dilution >= 0.0, '--dilution must be >= 0'
    assert args.minvolume >= 0.0, '--minvolume must be >= 0'
    assert args.maxvolume > 0.0, '--maxvolume must be > 0'
    # destination start
    args.destype = args.desttype.lower()
    if args.desttype == '96':
        destlimit = 96
    elif args.desttype == '384':
        destlimit = 384
    if args.deststart < 1 or args.deststart > destlimit:
        msg = 'Destination start well # must be in range: 1-{}'
        raise ValueError(msg.format(destlimit))
    # destination labware
    #args.destlabware = {x.split(':')[0]:x.split(':')[1] for x in args.destlabware.split(',')} 


def conc2df(concfile, row_select=None, file_format=None, header=True,
            labware_col=1, location_col=2, conc_col=3):
    """Loading a concentration file as a pandas dataframe
    """
    if header==True:
        header=0
    else:
        header=None
    # format
    if file_format is None:
        if concfile.endswith('.txt') or concfile.endswith('.csv'):
            file_format = 'tab'
        elif concfile.endswith('.xls') or concfile.endswith('.xlsx'):
            file_format = 'excel'
    else:
        file_format = file_format.lower()
        
    # load via pandas IO
    if file_format == 'tab':
        df = pd.read_csv(concfile, sep='\t', header=header)
    elif file_format == 'excel':
        xls = pd.ExcelFile(concfile)
        df = pd.read_excel(xls, header=header)
    else:
        raise ValueError('Concentration file not in usable format')

    # selecting particular rows
    if row_select is not None:
        df = df.iloc[row_select]

    # selecting columns
    df = df.iloc[:,[labware_col-1,location_col-1,conc_col-1]]
    df.columns = ['labware', 'location', 'conc']

    # return
    return df


def check_df_conc(df_conc, args):
    """Assertions of df_conc object formatting
    """
    # checking sample locations (>=1)
    msg = 'ERROR (concfile, line={}): location is < 1'
    for i,loc in enumerate(df_conc['location']):
        if loc < 1:
            print(msg.format(i), file=sys.stderr)
    
    # checking sample conc
    msg = 'ERROR (concfile, line={}): concentration is <= 0'
    for i,sc in enumerate(df_conc['conc']):
        if sc <= 0.0:
            print(msg.format(i), file=sys.stderr)


def dilution_volumes(df_conc, dilute_conc, min_vol, max_vol, 
                     min_total, dest_type='96'):
    """Setting the amoutn of sample to aliquot for dilution
    df_conc: pd.dataframe
    dilute_conc: concentration to dilute to 
    min_vol: min volume of sample to use
    max_vol: max total volume to use
    min_total: minimum total volume
    """
    # c1*v1 = c2*v2 
    # v2 = c1 * v1 / c2
    # v1 = c2 * v2 / c1

    # max well volume
    if dest_type == '96':
        max_well_vol = 280
    elif dest_type == '384':
        max_well_vol = 140
    else:
        msg = 'Destination labware type "{}" not recognized'
        raise ValueError(msg.format(dest_type))

    # range of dilutions
    samp_vol_range = max_vol - min_vol
    target_total_vol = round(samp_vol_range / 2 + min_vol)
    
    # bare min for sample volume used
    df_conc['total_volume'] = [x * min_vol / dilute_conc for x in df_conc['conc']]
    if max(df_conc['total_volume']) > max_well_vol:
        msg = 'ERROR: dilution volume exceeds max possible well volume.'
        msg += ' Lower --minvolume or chane destination labware type.'
        raise ValueError(msg)
    
    # raising total_vols if low volume (if small dilution factor)
    df_conc.loc[df_conc.total_volume < min_total, 'total_volume'] = min_total
    # setting volumes
    ## v1 = c2*v2/c1
    ## sample_volume = dilute_conc * total_volume / conc 
    f = lambda row: dilute_conc * row['total_volume'] / row['conc']
    df_conc['sample_volume'] = df_conc.apply(f, axis=1)
    # dilutatant volume = total_volume - sample_volume
    f = lambda row: row['total_volume'] - row['sample_volume']
    df_conc['dilutant_volume'] = df_conc.apply(f, axis=1)
    
    # return
    return df_conc
        

def add_dest(df_conc, dest_labware, dest_start=1):
    """Setting destination locations for samples & primers.
    Adding to df_conc:
      [dest_labware, dest_location]
    """
    dest_start= int(dest_start)
    #try:
    #    dest_labware = dest_labware_index[dest_type]
    #except KeyError:
    #    raise KeyError('Destination labware type not recognized')
    
    # adding columns
    df_conc['dest_labware'] = dest_labware
    dest_end = dest_start + df_conc.shape[0] 
    df_conc['dest_location'] = list(range(dest_start, dest_end))

    # return
    return df_conc


def reorder_384well(df, reorder_col):
    """Reorder values so that the odd, then the even locations are
    transferred. This is faster for a 384-well plate
    df: pandas.DataFrame
    reorder_col: column name to reorder
    """
    df['TECAN_sort_IS_EVEN'] = [x % 2 == 0 for x in df[reorder_col]]
    df.sort_values(by=['TECAN_sort_IS_EVEN', reorder_col], inplace=True)
    df = df.drop('TECAN_sort_IS_EVEN', 1)
    df.index = range(df.shape[0])
    return df


def pip_dilutant(df_conc, outFH, src_labware='100ml[001]'):
    """Writing worklist commands for aliquoting dilutant.
    Using 1-asp-multi-disp with 200 ul tips.
    Method:
    * calc max multi-dispense for 50 or 200 ul tips 
    """
    # determing how many multi-disp
    max_vol = max(df_conc.dilutant_volume)
    if max_vol > 180:
        n_disp = int(np.floor(900 / max_vol))  # using 1000 ul tips
    else:
        n_disp= int(np.floor(180 / max_vol))   # using 200 ul tips

    # making multi-disp object
    outFH.write('C;Dilutant\n')
    MD = Fluent.multi_disp()
    MD.SrcRackLabel = src_labware
    MD.SrcPosition = 1                               # need to set for all channels?
    MD.DestRackLabel = df_conc.dest_labware
    MD.DestPositions = df_conc.dest_location
    MD.Volume = df_conc.dilutant_volume             
    MD.NoOfMultiDisp = n_disp
    # writing
    outFH.write(MD.cmd() + '\n')


def pip_samples(df_conc, outFH):
    """Commands for aliquoting samples into dilutant
    """
    outFH.write('C;Samples\n')
    # for each Sample-PCR_rxn_rep, write out asp/dispense commands
    for i in range(df_conc.shape[0]):
        # aspiration
        asp = Fluent.aspirate()
        asp.RackLabel = df_conc.ix[i,'labware']
        asp.Position = df_conc.ix[i,'location']
        asp.Volume = round(df_conc.ix[i,'sample_volume'], 2)
        asp.LiquidClass = 'Water Contact Wet Single'
        outFH.write(asp.cmd() + '\n')

        # dispensing
        disp = Fluent.dispense()
        disp.RackLabel = df_conc.ix[i,'dest_labware']
        disp.Position = df_conc.ix[i,'dest_location']
        disp.Volume = round(df_conc.ix[i,'sample_volume'], 2)
        disp.LiquidClass = 'Water Contact Wet Single'
        outFH.write(disp.cmd() + '\n')

        # tip to waste
        outFH.write('W;\n')


def main(args=None):
    # Input
    if args is None:
        args = parse_args()
    check_args(args)
    # Import
    df_conc = conc2df(args.concfile, 
                      file_format=args.format,
                      row_select=args.rows, 
                      header=args.header,
                      labware_col=args.labware,
                      location_col=args.location, 
                      conc_col=args.conc)
    check_df_conc(df_conc, args)

    # Determining dilution volumes
    df_conc = dilution_volumes(df_conc, 
                               dilute_conc=args.dilution,
                               min_vol=args.minvolume,
                               max_vol=args.maxvolume,
                               min_total=args.mintotal,
                               dest_type=args.desttype)
    
    # Adding destination data
    df_conc = add_dest(df_conc, 
                       dest_labware=args.dest,
                       dest_start=args.deststart)
    
    # Reordering dest if plate type is 384-well
    if args.desttype == '384':
        df_conc = reorder_384well(df_conc, 'dest_location')
    elif args.desttype == '96':
        pass
    else:
        msg = 'Destination labware type "{}" not recognized'
        raise ValueError(msg.format(args.desttype))
    
    # Writing out gwl file
    gwl_file = args.prefix + '.gwl'
    with open(gwl_file, 'w') as gwlFH:
        ## Dilutant
        pip_dilutant(df_conc, outFH=gwlFH, src_labware=args.dlabware)
        ## Sample
        pip_samples(df_conc, outFH=gwlFH)

    # Writing out table
    conc_file = args.prefix + '_conc.txt'
    df_conc.round(1).to_csv(conc_file, sep='\t', index=False)

    # Create windows-line breaks formatted versions
    gwl_file_win = Utils.to_win(gwl_file)
    conc_file_win = Utils.to_win(conc_file)

    # end
    return (gwl_file, gwl_file_win, conc_file, conc_file_win)


# main
if __name__ == '__main__':
    pass





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
## batteries
import os
import sys
import unittest
## 3rd party
import pandas as pd
## package
from leylab_pipelines import Fluent


# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
class Test_Fluent(unittest.TestCase):

    def setUp(self):
        self.rd = Fluent.reagent_distribution()
        self.asp = Fluent.aspirate()
        self.disp = Fluent.dispense()

    def tearDown(self):
        pass

    def test_init_rd(self):
        cmd = self.rd.cmd()
        self.assertTrue(isinstance(cmd, str))

    def test_init_asp(self):
        cmd = self.asp.cmd()
        self.assertTrue(isinstance(cmd, str))

    def test_init_rd(self):
        cmd = self.disp.cmd()
        self.assertTrue(isinstance(cmd, str))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_leylab_pipelines
----------------------------------

Tests for `leylab_pipelines` module.
"""


import sys
import unittest

from leylab_pipelines import leylab_pipelines



class TestLeylab_pipelines(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_someting(self):
        pass



# -*- coding: utf-8 -*-
# import
## batteries
from __future__ import print_function
import os
import sys
import argparse
## package
### TECAN
from leylab_pipelines import Map2Robot
from leylab_pipelines import Dilute
from leylab_pipelines import QPCR
### LLP-DB
from leylab_pipelines import Convert
from leylab_pipelines import Acc2TaxID
from leylab_pipelines import TaxID2Lin
### LLP
from leylab_pipelines import Join


def TECAN_arg_parse():
  desc = 'Tools for working with the TECAN robot' 
  # subcommand descriptions
  epi = 'SUBCOMMANDS:\n'
  epi = epi + '  map2robot - ' + Map2Robot.get_desc() + '\n'
  epi = epi + '  dilute - ' + Dilute.get_desc() + '\n'
  epi = epi + '  qPCR - ' + QPCR.get_desc() + '\n'
  
  # main command arg parser
  parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                   formatter_class=argparse.RawTextHelpFormatter)
  
  # subparsers
  subparsers = parser.add_subparsers(dest='subparser_name')
  parser_map2robot = Map2Robot.parse_args(subparsers=subparsers)
  parser_dilute = Dilute.parse_args(subparsers=subparsers)
  parser_QPCR = QPCR.parse_args(subparsers=subparsers)

  # parsing args
  args = parser.parse_args()
  
  # running function
  if args.subparser_name == 'map2robot':
    Map2Robot.main(args)
  elif args.subparser_name == 'dilute':
    Dilute.main(args)
  elif args.subparser_name == 'qPCR':
    QPCR.main(args)
  else:
    msg = 'Command not recognized: "{}"'
    raise ValueError(msg.format(args.subparser_name))


def LLP_DB_arg_parse():
  desc = 'Tools for working with public databases' 
  # subcommand descriptions
  epi = 'SUBCOMMANDS:\n'
  epi = epi + '  convert - ' + Convert.get_desc() + '\n'
  epi = epi + '  acc2taxID - ' + Acc2TaxID.get_desc() + '\n'
  epi = epi + '  taxID2lin - ' + TaxID2Lin.get_desc() + '\n'
  
  # main command arg parser
  parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                   formatter_class=argparse.RawTextHelpFormatter)
  
  # subparsers
  subparsers = parser.add_subparsers(dest='subparser_name')
  parser_convert = Convert.parse_args(subparsers=subparsers)
  parser_taxID2lin = TaxID2Lin.parse_args(subparsers=subparsers)
  parser_acc2taxID = Acc2TaxID.parse_args(subparsers=subparsers)
  # parsing args
  args = parser.parse_args()
  
  # running function
  if args.subparser_name == 'convert':
    Convert.main(args)
  elif args.subparser_name == 'taxID2lin':
    TaxID2Lin.main(args)
  elif args.subparser_name == 'acc2taxID':
    Acc2TaxID.main(args)
  else:
    msg = 'Command not recognized: "{}"'
    raise ValueError(msg.format(args.subparser_name))


def LLP_arg_parse():
  desc = 'General bioinformatic tools'
  # subcommand descriptions
  epi = 'SUBCOMMANDS:\n'
  epi = epi + '  join - ' + Join.get_desc() + '\n'
  
  # main command arg parser
  parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                   formatter_class=argparse.RawTextHelpFormatter)
  
  # subparsers
  subparsers = parser.add_subparsers(dest='subparser_name')
  parser_join = Join.parse_args(subparsers=subparsers)
  # parsing args
  args = parser.parse_args()
  
  # running function
  if args.subparser_name == 'join':
    Join.main(args)
  else:
    msg = 'Command not recognized: "{}"'
    raise ValueError(msg.format(args.subparser_name))



# import
## batteries
import re
import os
import io
import sys
import gzip
import tempfile
import argparse
import logging
import urllib
## 3rd party
import numpy as np
import pandas as pd
## package
from leylab_pipelines import Utils 

# logging
logging.basicConfig(
    level=logging.DEBUG, format='%(asctime)s|%(levelname)s|%(message)s')

# functions
def get_desc():
    desc = 'Get NCBI taxonomy IDs from NCBI accessions'
    return desc

def parse_args(test_args=None, subparsers=None):
    # desc
    desc = get_desc()
    epi = """DESCRIPTION:

    TO CONVERT ACCESSION TO TAX_ID: 
      see ftp://ftp.ncbi.nih.gov/pub/taxonomy/accession2taxid/nucl_gb.accession2taxid.gz
    """
    if subparsers:
        parser = subparsers.add_parser('acc2taxID', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # args
    acc = parser.add_argument_group('Accessions')
    acc.add_argument('accessions', metavar='accessions', type=str,
                     help='Input table containing the acessions, "-" if from STDIN')
    acc.add_argument('-c', '--column', default=1,
                     help='Column number containing the accessions (default: %(default)s)')
    acc.add_argument('-s', '--sep', default='\t',
                     help='Column separator (default: %(default)s)')
    acc.add_argument('-x', '--no-header', default=False, action='store_true',
                     help='No header in input table (default: %(default)s)')
    acc.add_argument('-o', '--outfile', default='-',
                     help='Output file name; "-" if to STDOUT (default: %(default)s)')

    tax = parser.add_argument_group('Taxonomy database')
    tax.add_argument('-t', '--tax', default=None, nargs='+',
                     help='>=1 NCBI taxonomy file name (default: %(default)s)') 
    tax.add_argument('-u', '--url', default='ftp://ftp.ncbi.nih.gov/pub/taxonomy/accession2taxid/',
                     help='Base url for downloading the NCBI taxonomy files (default: %(default)s)')
    tax.add_argument('-T', '--types', default=['gb','wgs'], nargs='+',
                     help='>=1 DBs to download if no input files provided (default: %(default)s)') 
    tax.add_argument('-d', '--outdir', default=None,
                     help='Output directory for the taxonomy dump download. (default: %(default)s)')

    misc = parser.add_argument_group('Misc')
    misc.add_argument('-p', '--procs', default=1,
                     help='Number of processors to use. (default: %(default)s)')

    # running test args
    if test_args:
        args = parser.parse_args(test_args)
        return args


def load_acc_df(infile, sep='\t', no_header=False):
    logging.info('loading accessions...')

    if no_header is True:
        header = None
    else:
        header = 0
    if infile == '-':
        df = pd.read_csv(sys.stdin, sep=sep, header=header)
    else:
        df = pd.read_csv(infile, sep=sep, header=header)
    return df


def make_db_url(base_url, db):
    # urls for DB files to download
    DB_psbl = {'wgs' : 'nucl_wgs.accession2taxid.gz',
               'gb' : 'nucl_est.accession2taxid.gz',
               'est' : 'nucl_est.accession2taxid.gz',
               'gss' : 'nucl_gss.accession2taxid.gz'}
    
    try:
        x = DB_psbl[db]
    except KeyError:
        raise KeyError('DB type "{}" not recognized'.format(db))
    url = base_url + '/' + x 
    return [url, x]


def download_file(url, fileName, outDir=None):
    if outDir is None:
        outDir = tempfile.gettempdir()
    dbFile = os.path.join(outDir, fileName)

    if sys.version_info[0] >= 3:
        dbFile, headers = urllib.request.urlretrieve(url, filename=dbFile)
    else:
        dbFile, headers = urllib.urlretrieve(url, filename=dbFile)        

    logging.info('downloaded file: {}'.format(dbFile))
    return dbFile

    
def get_tax_db(base_url, DBs, outDir=None):
    """Getting NCBI DB file
    Saving to a temporary directory by default
    """
    logging.info('downloading NCBI taxonomy dump...')
    
    # url(s)
    urls = [make_db_url(base_url, x) for x in DBs]

    # downloading
    dmpFiles = [download_file(url[0], url[1], outDir) for url in urls]
                
    ## checking for existence
    for F in dmpFiles:
        if not os.path.isfile(F):
            raise ValueError('Cannot find file: {}'.format(F))

    return dmpFiles
    

def acc_to_taxID(db_file, df_acc, column=1):
    logging.info('mapping accessions with file: {}'.format(db_file))

    # accs 
    column = int(column) - 1
    accs = df_acc.iloc[:,column].tolist()
    # determining taxonomic IDs
    if db_file.endswith('.gz'):
        inF = gzip.open(db_file, 'rt') 
    else:
        inF =  open(db_file, 'r') 
    taxIDs = np.empty(df_acc.shape[0], dtype=object)
    accs_len = len(accs)
    for i,line in enumerate(inF):
        if (i+1) % 1000000 == 0:
            logging.info('Number of DB records processed: {}'.format(i+1))
        line = line.rstrip().split('\t')
        for i in range(accs_len):
            if accs[i] == line[0]:
                taxIDs[i] = line[2]
                continue

    return taxIDs


def which_taxID(x):
    for y in x:
        if y is not None:
            return y
    return None
        

def write_table(df, outFile, no_header, sep='\t'):
    header = no_header == False
    if outFile == '-':
        outFile = sys.stdout
    df.to_csv(outFile, sep=sep, header=header, index=False)


def main(args=None):
    # Input
    if args is None:
        args = parse_args()

    # load accessions
    df_acc = load_acc_df(args.accessions, args.sep, no_header=args.no_header)

    # data downloaded from ftp://ftp.ncbi.nih.gov/pub/taxonomy/
    if args.outdir is not None or args.tax is None:
        args.tax = get_tax_db(args.url, args.types, args.outdir)

    # getting taxIDs for accessions
    taxIDs = [acc_to_taxID(db_file, df_acc, column=args.column) for db_file in args.tax]
    taxIDs = np.array(taxIDs)

    # taking not None taxID
    df_acc['TaxID'] = np.apply_along_axis(which_taxID, 0, taxIDs) 

    # writing out file
    write_table(df_acc, outFile=args.outfile, no_header=args.no_header, sep=args.sep)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
## batteries
import os
import sys
import unittest
## package
from leylab_pipelines import Convert


# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')

# global variables
ENSEMBL_ID = 'ENST00000407559'

HGNC_ID = '9245'
ENTREZ_ID = '39'
UNIPROT_ID = 'Q9BWD1'
ACCESSION_ID = 'AC131209'

EMAIL = 'dummy@dummy.info'
FAKE_EMAIL = 'pyentrez.info'


# tests
class Test_Convert(unittest.TestCase):

    def setUp(self):
        self.Id = Convert.Conversion(EMAIL)

    def tearDown(self):
        pass

    def test_email(self):
        self.assertRaises(ValueError, Convert.Conversion, FAKE_EMAIL)

    def test_convert_ensembl_to_entrez(self):
        # real ID
        ID = self.Id.convert_ensembl_to_entrez(ENSEMBL_ID)
        self.assertListEqual(ID, [ENSEMBL_ID, '55112'])
        # fake ID
        ID = self.Id.convert_ensembl_to_entrez(ENSEMBL_ID + '_FAKEID')
        self.assertListEqual(ID, [ENSEMBL_ID + '_FAKEID', None])
        
    def test_convert_hgnc_to_entrez(self):
        # real ID
        ID = self.Id.convert_hgnc_to_entrez(HGNC_ID)
        self.assertListEqual(ID, [HGNC_ID, '8500'])
        # fake ID
        ID = self.Id.convert_hgnc_to_entrez(HGNC_ID + '_FAKEID')
        self.assertListEqual(ID, [HGNC_ID + '_FAKEID', None])        

    def test_convert_entrez_to_uniprot(self):
        # real ID
        ID = self.Id.convert_entrez_to_uniprot(ENTREZ_ID)
        self.assertListEqual(ID, [ENTREZ_ID, 'Q9BWD1'])
        # fake ID

    def test_convert_uniprot_to_entrez(self):
        # real ID
        ID = self.Id.convert_uniprot_to_entrez(UNIPROT_ID)
        self.assertListEqual(ID, [UNIPROT_ID, '39'])
        # fake ID
        

    def test_convert_accesion_to_taxid(self):
        # real ID
        ID = self.Id.convert_accession_to_taxid(ACCESSION_ID)
        self.assertListEqual(ID, [ACCESSION_ID, '9606'])
        # fake ID
        ID = self.Id.convert_accession_to_taxid(ACCESSION_ID + '_FAKEID')
        self.assertListEqual(ID, [ACCESSION_ID + '_FAKEID', None])
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
## batteries
import os
import sys
import unittest
## 3rd party
import pandas as pd
## package
from leylab_pipelines import Map2Robot
from leylab_pipelines import Utils


# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
class Test_Map2Robot_parser(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    # args
    def test_parser_dest(self):
        # 96-well
        args = Map2Robot.parse_args(['--desttype', '96',
                                     '--deststart', '97',
                                     'mapfile'])
        with self.assertRaises(ValueError):
            Map2Robot.check_args(args)
        # 384-well
        args = Map2Robot.parse_args(['--desttype', '384',
                                     '--deststart', '389',
                                     'mapfile'])
        with self.assertRaises(ValueError):
            Map2Robot.check_args(args)

    def test_parser_tube(self):
        # mastermix <1
        args = Map2Robot.parse_args(['--mmtube', '0', 'mapfile'])
        with self.assertRaises(ValueError):
            Map2Robot.check_args(args)
        # mastermix >24
        args = Map2Robot.parse_args(['--mmtube', '25', 'mapfile'])
        with self.assertRaises(ValueError):
            Map2Robot.check_args(args)


class Test_Map2Robot_import(unittest.TestCase):

    def setUp(self):
        mapfile = os.path.join(data_dir, 'mapping_file_fecal_stability.txt')
        args = Map2Robot.parse_args([mapfile])
        Map2Robot.check_args(args)
        self.df_map = Map2Robot.map2df(args.mapfile)

    def tearDown(self):
        self.df_map = None
        pass

    # import
    def test_load_map_txt(self):
        self.assertTrue(isinstance(self.df_map, pd.DataFrame))

    def test_load_map_xls(self):
        self.assertTrue(isinstance(self.df_map, pd.DataFrame))


class Test_Map2Robot_rows(unittest.TestCase):

    def setUp(self):
        mapfile = os.path.join(data_dir, 'mapping_file_fecal_stability.txt')
        self.args = Map2Robot.parse_args(['--rows', '1,2,4-5', mapfile])
        Map2Robot.check_args(self.args)
        self.df_map = Map2Robot.map2df(self.args.mapfile,
                                       row_select=self.args.rows)

    def tearDown(self):
        self.df_map = None
        pass

    # import
    def test_load_map_txt(self):
        self.assertTrue(isinstance(self.df_map, pd.DataFrame))
        self.assertEqual(self.df_map.shape[0], 4)
        self.assertListEqual(list(self.df_map.index), [0,1,3,4])

class Test_Map2Robot_addDest(unittest.TestCase):

    def setUp(self):
        mapfile = os.path.join(data_dir, 'mapping_file_fecal_stability.txt')
        self.args = Map2Robot.parse_args([mapfile])
        Map2Robot.check_args(self.args)
        self.df_map = Map2Robot.map2df(self.args.mapfile)

    def tearDown(self):
        self.df_map = None
        self.args = None

    # adding destination
    def test_load_map_txt(self):
        self.df_map = Map2Robot.add_dest(self.df_map, self.args.dest)
        self.assertTrue(isinstance(self.df_map, pd.DataFrame))

    # destination start offset
    def test_load_map_deststart(self):
        self.df_map = Map2Robot.add_dest(self.df_map,
                                         self.args.dest,
                                         dest_start=49)

        self.assertTrue(isinstance(self.df_map, pd.DataFrame))
        loc_start = self.df_map.loc[0,'TECAN_dest_location']
        self.assertEqual(loc_start, 49)


class Test_Map2Robot_destStart384(unittest.TestCase):

    def setUp(self):
        mapfile = os.path.join(data_dir, 'mapping_file_fecal_stability.txt')
        self.args = Map2Robot.parse_args(['--desttype', '384', mapfile])
        Map2Robot.check_args(self.args)
        self.df_map = Map2Robot.map2df(self.args.mapfile)

    def tearDown(self):
        self.df_map = None
        self.args = None

    def test_load_map_deststart384(self):
        self.df_map = Map2Robot.add_dest(self.df_map,
                                         self.args.dest,
                                         dest_type=self.args.desttype,
                                         dest_start=200)
        self.assertTrue(isinstance(self.df_map, pd.DataFrame))
        loc_start = self.df_map.ix[0,'TECAN_dest_location']
        self.assertEqual(loc_start, 200.0)

    def test_load_map_deststart384_2(self):
        self.df_map = Map2Robot.add_dest(self.df_map,
                                         self.args.dest,
                                         dest_type=self.args.desttype,
                                         dest_start=370)
        self.assertTrue(isinstance(self.df_map, pd.DataFrame))

        loc_start = self.df_map.ix[0,'TECAN_dest_location']
        self.assertEqual(loc_start, 370.0)

        i = self.df_map.shape[0]-1
        loc_end = self.df_map.ix[i,'TECAN_dest_location']
        self.assertEqual(loc_end, 384.0)

    # destination location reorder for 384-well plates
    def test_load_map_destreorder(self):
        self.df_map = Map2Robot.add_dest(self.df_map, self.args.dest)
        self.df_map = Map2Robot.reorder_384well(self.df_map, 'TECAN_dest_location')

        self.assertTrue(isinstance(self.df_map, pd.DataFrame))
        loc_start = self.df_map.loc[0,'TECAN_dest_location']
        self.assertEqual(loc_start, 1)
        nrow = self.df_map.shape[0]
        loc_end = self.df_map.loc[nrow-1,'TECAN_dest_location']
        self.assertEqual(loc_end, 96)


class Test_Map2Robot_pipMasterMix(unittest.TestCase):

    def setUp(self):
        mapfile = os.path.join(data_dir, 'mapping_file_fecal_stability.txt')
        self.args = Map2Robot.parse_args([mapfile])
        Map2Robot.check_args(self.args)
        self.df_map = Map2Robot.map2df(self.args.mapfile)
        self.df_map = Map2Robot.add_dest(self.df_map,
                                         self.args.dest,
                                         dest_type=self.args.desttype)
        if self.args.desttype == '384':
            self.df_map = reorder_384well(self.df_map, 'TECAN_dest_location')
        self.gwl = '/tmp/MM.gwl'
        self.gwlFH = open(self.gwl, 'w')

    def tearDown(self):
        self.df_map = None
        self.args = None

    def test_pip_master_mix(self):
        Map2Robot.pip_mastermix(self.df_map, self.gwlFH)
        self.gwlFH.close()
        ret = Utils.check_gwl(self.gwl)
        self.assertIsNone(ret)


class Test_Map2Robot_main(unittest.TestCase):

    def setUp(self):
        mapfile = os.path.join(data_dir, 'mapping_file_fecal_stability.txt')
        self.args = Map2Robot.parse_args(['--prefix', '/tmp/MAP', mapfile])
        self.files = Map2Robot.main(self.args)

    def tearDown(self):
        pass

    def test_main_gwl(self):
        ret = Utils.check_gwl(self.files[0])
        self.assertIsNone(ret)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import glob

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'numpy',
    'pandas',
    'dask',
    'xmltodict'
]

test_requirements = [
    # TODO: put package test requirements here
]

scripts = ['scripts/TECAN', 'scripts/LLP-DB', 'scripts/LLP']

long_desc = """
General bioinformatic pipelines associated with the Ley Lab at the MPI in Tuebingen.
For more information, see the README.
"""


setup(
    name='leylab_pipelines',
    version='0.2.0',
    description="General bioinformatic pipelines associated with the Ley Lab at the MPI in Tuebingen",
    long_description=long_desc + '\n\n' + history,
    author="Nick Youngblut",
    author_email='leylabmpi@gmail.com',
    url='https://github.com/leylabmpi/leylab_pipelines',
    packages=[
        'leylab_pipelines',
    ],
    package_dir={'leylab_pipelines':
                 'leylab_pipelines'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='leylab_pipelines',
    classifiers=[
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    scripts=scripts
)



# import
## batteries
import re
import os
import sys
import gzip
import tempfile
import multiprocessing
import argparse
import logging
import urllib
import tarfile
## 3rd party
import pandas as pd
## package
from leylab_pipelines import Utils 

# logging
logging.basicConfig(
    level=logging.DEBUG, format='%(asctime)s|%(levelname)s|%(message)s')

# functions
def get_desc():
    desc = 'Convert an NCBI taxonomy (taxdump) into lineages (taxID<-->lineage)'
    return desc

def parse_args(test_args=None, subparsers=None):
    # desc
    desc = get_desc()
    epi = """DESCRIPTION:
    Convert the NCBI taxonomy dump to a lineage table.
    The lineage table can be used to map taxonomy IDs to taxonomy. 

    If the specific taxonomy dump files are not provided, this command
    will download the full dump file, uncompress it, and use the needed
    files. By default, the dump files are downloaded to a temporary directory.
    You can keep the taxonomy dump by using the --outdir option.

    TO CONVERT ACCESSION TO TAX_ID: 
      see ftp://ftp.ncbi.nih.gov/pub/taxonomy/accession2taxid/nucl_gb.accession2taxid.gz
    """
    if subparsers:
        parser = subparsers.add_parser('taxID2lin', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # args
    io = parser.add_argument_group('File input/output')
    io.add_argument('--nodes', default=None,
                    help='nodes.dmp file path. (default: %(default)s)')
    io.add_argument('--names', default=None,
                    help='names.dmp file path. (default: %(default)s)')
    io.add_argument('-o', '--outfile', default='NCBI_taxID2lin.txt',
                    help='Output file for lineage table (default: %(default)s)')
    
    dmp = parser.add_argument_group('Taxonomy dump')
    dmp.add_argument('-u', '--url', default='ftp://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz',
                    help='URL for downloading taxonomy dump. (default: %(default)s)')
    dmp.add_argument('-d', '--outdir', default=None,
                     help='Output directory for the taxonomy dump download. (default: %(default)s)')
    dmp.add_argument('-a', '--all', action='store_true', default=False,
                     help='Output all taxonomy levels instead of just the standard levels. (default: %(default)s)')

    misc = parser.add_argument_group('Misc')
    misc.add_argument('-p', '--procs', default=1,
                      help='Number of processors to use. (default: %(default)s)')

    # running test args
    if test_args:
        args = parser.parse_args(test_args)
        return args


def strip(str_):
    '''
    :param str_: a string
    '''
    return str_.strip()


def load_nodes(nodes_file):
    '''
    load nodes.dmp and convert it into a pandas.DataFrame
    '''
    df = pd.read_csv(nodes_file, sep='|', header=None, index_col=False,
                     names=[
                         'tax_id',
                         'parent_tax_id',
                         'rank',
                         'embl_code',
                         'division_id',
                         'inherited_div_flag',
                         'genetic_code_id',
                         'inherited_GC__flag',
                         'mitochondrial_genetic_code_id',
                         'inherited_MGC_flag',
                         'GenBank_hidden_flag',
                         'hidden_subtree_root_flag',
                         'comments'
                     ])

    # To get rid of flanking tab characters
    df['rank'] = df['rank'].apply(strip)
    df['embl_code'] = df['embl_code'].apply(strip)
    df['comments'] = df['comments'].apply(strip)
    return df


def load_names(names_file):
    '''
    load names.dmp and convert it into a pandas.DataFrame
    '''
    df = pd.read_csv(names_file, sep='|', header=None, index_col=False,
                     names=[
                         'tax_id',
                         'name_txt',
                         'unique_name',
                         'name_class'
                     ])
    df['name_txt'] = df['name_txt'].apply(strip)
    df['unique_name'] = df['unique_name'].apply(strip)
    df['name_class'] = df['name_class'].apply(strip)

    sci_df = df[df['name_class'] == 'scientific name']
    sci_df.reset_index(drop=True, inplace=True)
    return sci_df


def to_dict(lineage):
    """
    convert the lineage into a list of tuples in the form of

    [
        (tax_id1, rank1, name_txt1),
        (tax_id2, rank2, name_txt2),
        ...
    ]

    to a dict
    """
    dd = {}
    num_re = re.compile('[0-9]+')
    len_lineage = len(lineage)
    for k, __ in enumerate(lineage):
        tax_id, rank, name_txt = __
        # use the last rank as the tax_id, whatever it is, genus or species.
        if k == len_lineage - 1:
            dd['tax_id'] = tax_id

        # e.g. there could be multiple 'no rank'
        numbered_rank = rank
        while numbered_rank in dd:
            # print __, numbered_rank
            search = num_re.search(numbered_rank)
            if search is None:
                count = 1
            else:
                count = int(search.group()) + 1
            numbered_rank = '{0}{1}'.format(rank, count)
        dd[numbered_rank] = name_txt
    return dd


def find_lineage(tax_id):
    if tax_id % 50000 == 0:
        logging.debug('working on tax_id: {0}'.format(tax_id))
    lineage = []
    while True:
        rec = TAXONOMY_DICT[tax_id]
        lineage.append((rec['tax_id'], rec['rank'], rec['name_txt']))
        tax_id = rec['parent_tax_id']

        if tax_id == 1:
            break

    # reverse results in lineage of Kingdom => species, this is helpful for
    # to_dict when there are multiple "no rank"s
    lineage.reverse()
    return to_dict(lineage)


def get_taxdump(url, outDir=None):
    """Getting taxdump file from NCBI. 
    Saving to a temporary directory. 
    """
    logging.info('downloading NCBI taxonomy dump...')
    
    # downloading
    if outDir:
        # saving taxdump
        dmpFile = os.path.join(outDir, 'taxdump.tar.gz')
    else:
        # taxdump written to temporary directory
        outDir = tempfile.gettempdir()
        dmpFile = None
    if sys.version_info[0] >= 3:
        dmpFile, headers = urllib.request.urlretrieve(url, filename=dmpFile)
    else:
        dmpFile, headers = urllib.urlretrieve(url, filename=dmpFile)        
            
    # uncompressing
    logging.info('uncompressing NCBI taxonomy dump file...')
    tar = tarfile.open(dmpFile, "r:gz")
    tar.extractall(path=outDir)
    tar.close()
    logging.info('files uncompressed to: {}'.format(outDir))

    # target file names
    files = {'nodes' : os.path.join(outDir, 'nodes.dmp'),
             'names' : os.path.join(outDir, 'names.dmp')}
    ## checking for existence
    for k,v in files.items():
        if not os.path.isfile(v):
            raise ValueError('Cannot find file: {}'.format(v))
    # ret
    return files
    

def main(args=None):
    # Input
    if args is None:
        args = parse_args()

    # data downloaded from ftp://ftp.ncbi.nih.gov/pub/taxonomy/
    if args.outdir is not None or args.nodes is None or args.names is None:
        files = get_taxdump(args.url, args.outdir)
        args.nodes = files['nodes']
        args.names = files['names']
        
    nodes_df = load_nodes(args.nodes)
    names_df = load_names(args.names)
    df = nodes_df.merge(names_df, on='tax_id')
    df = df[['tax_id', 'parent_tax_id', 'rank', 'name_txt']]
    df.reset_index(drop=True, inplace=True)
    logging.info('# of tax ids: {0}'.format(df.shape[0]))
    # log summary info about the dataframe
    df.info()

    # force to use global variable TAXONOMY_DICT because map doesn't allow
    # passing extra args easily
    global TAXONOMY_DICT
    logging.info('generating TAXONOMY_DICT...')
    TAXONOMY_DICT = dict(zip(df.tax_id.values, df.to_dict('records')))

    ncpus = int(args.procs)
    logging.info('using {0} cpus to find lineages for all tax ids'.format(ncpus))
    pool = multiprocessing.Pool(ncpus)
    lineages_dd = pool.map(find_lineage, df.tax_id.values)
    pool.close()

    logging.info('generating a dictionary of lineages information...')
    dd_for_df = dict(zip(range(len(lineages_dd)), lineages_dd))

    logging.info('generating lineages_df...')
    lineages_df = pd.DataFrame.from_dict(dd_for_df, orient='index')
    lineages_df.sort_values('tax_id', inplace=True)

    logging.info('writing lineages to: {}'.format(args.outfile))
    cols = ['tax_id',
            'superkingdom',
            'phylum',
            'class',
            'order',
            'family',
            'genus',
            'species']
     
    other_cols = sorted([__ for __ in lineages_df.columns if __ not in cols])
    if args.all:
        output_cols = cols + other_cols
    else:
        output_cols = cols
    lineages_df.to_csv(args.outfile, index=False, columns=output_cols, sep='\t')




# import
## batteries
import os
import sys
import glob
import tempfile
import argparse
import logging
import shutil
## 3rd party
import dask.dataframe as dd
## package
from leylab_pipelines import Utils 

# logging
logging.basicConfig(
    level=logging.DEBUG, format='%(asctime)s|%(levelname)s|%(message)s')

# functions
def get_desc():
    desc = 'Fast table join (in parallel)'
    return desc

def parse_args(test_args=None, subparsers=None):
    # desc
    desc = get_desc()
    epi = """DESCRIPTION:
    Join tables with the dask python package, which enables parallel processing
    and fast execution.
    
    JOIN:
      To designate columns to join on, the --join input should be in the format:
      'X=X,Y=Y,...'
      For example, to join on 'accession' in table 1 with 'Acc' in table 2:
         --join 'accession=Acc'
      This can be extended to multiple column joins with a comma-sep list.

    DTYPEs:
      The -L and -R options designate the dtypes (eg. 'object' or 'float')
      for the columns that are joined (see JOIN). The dtypes MUST match for
      the matched joined columns. If only 1 dtype is provided for multiple 
      join columns, then that dtype is applied to all join columns.

    OUTPUT:
      The table partitions will be written to separate temporary files, then
      joined into the final output file (designated with --outfile)
    """
    if subparsers:
        parser = subparsers.add_parser('join', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # args
    io = parser.add_argument_group('Input/Output')
    io.add_argument('table', metavar='table', type=str, nargs=2,
                     help='File names of the tables to join on')
    io.add_argument('-o', '--outfile', default='-',
                     help='Output file name; "-" if to STDOUT (default: %(default)s)')
    jn = parser.add_argument_group('How to join')
    jn.add_argument('-j', '--join', default='1=1',
                     help='Columns to join on. See DESCRIPTION (default: %(default)s)')
    jn.add_argument('-H', '--how', default='inner',
                     help='How to join the tables (default: %(default)s)')                    
    jn.add_argument('-s', '--sep', default='\t',
                     help='Column separator (default: %(default)s)')
    jn.add_argument('-L', '--dtypeL', default='str',
                     help='dtype(s) for left table join column(s) (default: %(default)s)')    
    jn.add_argument('-R', '--dtypeR', default='str',
                     help='dtype(s) for right table join column(s) (default: %(default)s)')    

    misc = parser.add_argument_group('Misc')
    misc.add_argument('-c', '--clean', action='store_false', default=True,
                      help='Remove *.partd files from the temporary dir. (default: %(default)s)')

    # running test args
    if test_args:
        args = parser.parse_args(test_args)
        return args
        

def parse_join(join_str):
    """Parsing joing string in form of 'X=X,Y=Y'
    """
    join_on = {}
    for x in join_str.split(','):
        y = x.split('=')
        if len(y) != 2:
            msg = '--join should be in format "X=X,Y=Y"'
        try:
            join_on['left'].append(y[0])
        except KeyError:
            join_on['left'] = [y[0]]
        try:
            join_on['right'].append(y[1])
        except KeyError:
            join_on['right'] = [y[1]]

    return(join_on)


def parse_dtype(dtype_str, join_on):
    dtypes = dtype_str.split(',')
    if len(dtypes) == 1:
        dtypes = {x:dtype_str for x in join_on}
    elif len(dtypes) != len(join_on):
        msg = 'The number of dtypes should be 1 or the same length as --join'
        raise ValueError(msg)
    else:
        dtypes = {x[i]:dtypes[i] for i in range(len(dtypes))}
    return dtypes


def get_table(infile, sep='\t', dtype=None):
    if infile.endswith('.gz'):
        compression = 'gzip'
    else:
        compression = None
    df = dd.read_csv(infile, sep=sep, dtype=dtype, compression=compression)
    return(df)


def write_tmp(df):
    # writing out temporary files
    tmpDir = tempfile.gettempdir()
    tempFile = next(tempfile._get_candidate_names())
    tempFile = os.path.join(tmpDir, tempFile + '*.txt')
    logging.info('writing temporary files to: {}'.format(tempFile))
    df.to_csv(tempFile, index=False, sep='\t')
    return tempFile


def write_table(infile, outfile, conn='w', header=True):
    if outfile == '-':
        with open(infile, 'r') as inF:
            if header == False:
                line = inF.readline()
            for line in inF:
                sys.stdout.write(line)
    else:
        with open(infile, 'r') as inF, open(outfile, conn) as outF:
            if header == False:
                line = inF.readline()
            for line in inF:
                outF.write(line)

def clean_partd(tmpDir=None):
    if tmpDir is None:
        tmpDir = tempfile.gettempdir()    
    logging.info('Cleaning *.partd directories in : {}'.format(tmpDir))

    g = os.path.join(tmpDir, 'tmp*.partd')
    partd_dirs = glob.glob(g)
    for D in partd_dirs:
        try:
            shutil.rmtree(D)
        except PermissionError:
            pass

def main(args=None):
    # Input
    if args is None:
        args = parse_args()

    # clean tmp directory
    if args.clean:
        clean_partd()
 
    # parsing the join arg
    join_on = parse_join(args.join)
    ## parsing dtypes
    args.dtypeL = parse_dtype(args.dtypeL, join_on['left'])
    args.dtypeR = parse_dtype(args.dtypeR, join_on['right'])

    # creating table objects
    df1 = get_table(args.table[0], sep=args.sep, dtype=args.dtypeL)
    df2 = get_table(args.table[1], sep=args.sep, dtype=args.dtypeR)

    # joining (merging)
    df = dd.merge(df1, df2, left_on=join_on['left'], right_on=join_on['right'], how=args.how) #, npartitions=args.procs)
    
    # writing out temporary files
    tempFile_str = write_tmp(df)
    
    # cat files
    for i,infile in enumerate(glob.glob(tempFile_str)):
        if i == 0:
            conn = 'w'
            header = True
        else:
            conn = 'a'
            header = False        
        write_table(infile, args.outfile, conn=conn, header=header)
            
        



# -*- coding: utf-8 -*-

# import
from __future__ import print_function
import os
import sys
import logging
import numpy as np

# functions
def make_range(x, set_zero_index=False):
    """Making a range from a string of comma-delimited and hyphenated number strings.
    If x = 'all', None returned
    x : string
    set_zero_index : change from 1-index to 0-index
    Example: 1,2,3,4
    Example: 1,2,5-6
    """
    x = str(x)
    if x.lower() == 'all':
        return None
    y = str(x).split(',')

    z = []
    for i in y:
        j = [int(x) for x in i.split('-')]
        if len(j) > 1:
            z = z + list(range(j[0], j[1]+1))
        else:
            z = z + j

    if set_zero_index:
        z = [x-1 for x in z]
        if min(z) < 0:
            raise ValueError('Negative values from setting zero index')

    return z


def check_gwl(gwl_file):
    """Checking that gwl in correct format
    gwl_file : input file name or file handle
    """
    # file read
    try:
        gwl_file = open(gwl_file, 'r')
    except ValueError:
        pass
    # checks
    for line in gwl_file:
        line = line.rstrip()
        if line is '':
            msg = 'Empty lines not allowed in gwl files'
            raise ValueError(msg)
        if not line[0] in ('A', 'D', 'R', 'W', 'F', 'C', 'B'):
            msg = '"{}" not a valid command ID'
            raise ValueError(msg.format(line[0]))
    # file close
    gwl_file.close()


def to_win(file_name, suffix='_win'):
    """Create a copy of a file but with windows line breakds
    file_name : str, name of file
    suffix : added to file name of copy
    Returns : name of new file
    """
    x = os.path.splitext(file_name)
    out_file = x[0] + suffix + x[1]
    with open(file_name) as inFH, open(out_file, 'w') as outFH:
        for line in inFH:
            line = line.replace('\n', '\r\n')
            outFH.write(line)
    return out_file


def backup_file(f):
    """
    Back up a file, old_file will be renamed to #old_file.n#, where n is a
    number incremented each time a backup takes place
    """
    if os.path.exists(f):
        dirname = os.path.dirname(f)
        basename = os.path.basename(f)
        count = 1
        rn_to = os.path.join(
            dirname, '#' + basename + '.{0}#'.format(count))
        while os.path.exists(rn_to):
            count += 1
            rn_to = os.path.join(
                dirname, '#' + basename + '.{0}#'.format(count))
        logging.info("Backing up {0} to {1}".format(f, rn_to))
        os.rename(f, rn_to)
        return rn_to
    else:
        logging.warning('{0} doesn\'t exist'.format(f))

# main
if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import
## batteries
import os
import sys
import unittest
## 3rd party
import pandas as pd
## package
from leylab_pipelines import QPCR
from leylab_pipelines import Utils


# data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')



# tests
class Test_import(unittest.TestCase):

    def setUp(self):
        infile = os.path.join(data_dir, 'qPCR_setup/qPCR_Zach_plate1.xlsx')
        args = QPCR.parse_args([infile])
        QPCR.check_args(args)
        self.df_setup = QPCR.load_setup(args.setup)
        QPCR.check_df_setup(self.df_setup)

    def tearDown(self):
        self.df_setup = None
        pass

    # import
    def test_load_excel(self):
        self.assertTrue(isinstance(self.df_setup, pd.DataFrame))
        #print(self.df_setup)


class Test_add_dest(unittest.TestCase):

    def setUp(self):
        infile = os.path.join(data_dir, 'qPCR_setup/qPCR_Zach_plate1.xlsx')
        args = QPCR.parse_args([infile])
        QPCR.check_args(args)
        self.df_setup = QPCR.load_setup(args.setup)
        QPCR.check_df_setup(self.df_setup)
        QPCR.add_dest(self.df_setup, args.dest, dest_type=args.desttype)
        
    def tearDown(self):
        self.df_setup = None
        pass

    # import
    def test_add_dest(self):
        self.assertTrue(isinstance(self.df_setup, pd.DataFrame))
        #print(self.df_setup)

class Test_qPCR_main1(unittest.TestCase):

    def setUp(self):
        infile = os.path.join(data_dir, 'qPCR_setup/qPCR_Zach_plate1.xlsx')
        args = QPCR.parse_args(['--prefix', '/tmp/qPCR', infile])
        self.files = QPCR.main(args)

    def tearDown(self):
        pass

    def test_main_gwl(self):
        ret = Utils.check_gwl(self.files[0])
        self.assertIsNone(ret)
